#!/usr/bin/python3

# Copyright (c) 2015 Red Hat, Inc. <http://www.redhat.com/>
# This file is part of GlusterFS.
#
# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.

import sys
import os
import xattr
import uuid
import re
import errno

CHANGELOG_SEARCH_MAX_TRY = 31
DEC_CTIME_START = 5
ROOT_GFID = "00000000-0000-0000-0000-000000000001"
MAX_NUM_CHANGELOGS_TRY = 2


def output_not_found(gfid):
    # Write GFID to stderr
    sys.stderr.write("%s\n" % gfid)


def output_success(path):
    # Write converted Path to Stdout
    sys.stdout.write("%s\n" % path)


def full_dir_path(gfid):
    out_path = ""
    while True:
        path = os.path.join(".glusterfs", gfid[0:2], gfid[2:4], gfid)
        path_readlink = os.readlink(path)
        pgfid = os.path.dirname(path_readlink)
        out_path = os.path.join(os.path.basename(path_readlink), out_path)
        if pgfid == "../../00/00/%s" % ROOT_GFID:
            out_path = os.path.join("./", out_path)
            break
        gfid = os.path.basename(pgfid)
    return out_path


def find_path_from_changelog(fd, gfid):
    """
    In given Changelog File, finds using following pattern
    <T><GFID>\x00<TYPE>\x00<MODE>\x00<UID>\x00<GID>\x00<PARGFID>/<BASENAME>
    Pattern search finds PARGFID and BASENAME, Convert PARGFID to Path
    Using readlink and add basename to form Full path.
    """
    content = fd.read()

    pattern = "E%s" % gfid
    pattern += "\x00(3|23)\x00\d+\x00\d+\x00\d+\x00([^\x00]+)/([^\x00]+)"
    pat = re.compile(pattern)
    match = pat.search(content)

    if match:
        pgfid = match.group(2)
        basename = match.group(3)
        if pgfid == ROOT_GFID:
            return os.path.join("./", basename)
        else:
            full_path_parent = full_dir_path(pgfid)
            if full_path_parent:
                return os.path.join(full_path_parent, basename)

    return None


def gfid_to_path(gfid):
    """
    Try readlink, if it is directory it succeeds.
    Get ctime of the GFID file, Decrement by 5 sec
    Search for Changelog filename, Since Changelog file generated
    every 15 sec, Search and get immediate next Changelog after the file
    Creation. Get the Path by searching in Changelog file.
    Get the resultant file's GFID and Compare with the input, If these
    GFIDs are different then Some thing is changed(May be Rename)
    """
    gfid = gfid.strip()
    gpath = os.path.join(".glusterfs", gfid[0:2], gfid[2:4], gfid)
    try:
        output_success(full_dir_path(gfid))
        return
    except OSError:
        # Not an SymLink
        pass

    try:
        ctime = int(os.stat(gpath).st_ctime)
        ctime -= DEC_CTIME_START
    except (OSError, IOError):
        output_not_found(gfid)
        return

    path = None
    found_changelog = False
    changelog_parse_try = 0
    for i in range(CHANGELOG_SEARCH_MAX_TRY):
        cl = os.path.join(".glusterfs/changelogs", "CHANGELOG.%s" % ctime)

        try:
            with open(cl, "rb") as f:
                changelog_parse_try += 1
                found_changelog = True
                path = find_path_from_changelog(f, gfid)
                if not path and changelog_parse_try < MAX_NUM_CHANGELOGS_TRY:
                    ctime += 1
                    continue
            break
        except (IOError, OSError) as e:
            if e.errno == errno.ENOENT:
                ctime += 1
            else:
                break

    if not found_changelog:
        output_not_found(gfid)
        return

    if not path:
        output_not_found(gfid)
        return
    gfid1 = str(uuid.UUID(bytes=xattr.get(path, "trusted.gfid")))
    if gfid != gfid1:
        output_not_found(gfid)
        return

    output_success(path)


def main():
    num_arguments = 3
    if not sys.stdin.isatty():
        num_arguments = 2

    if len(sys.argv) != num_arguments:
        sys.stderr.write("Invalid arguments\nUsage: "
                         "%s <BRICK_PATH> <GFID_FILE>\n" % sys.argv[0])
        sys.exit(1)

    path = sys.argv[1]

    if sys.stdin.isatty():
        gfid_list = os.path.abspath(sys.argv[2])
        os.chdir(path)
        with open(gfid_list) as f:
            for gfid in f:
                gfid_to_path(gfid)
    else:
        os.chdir(path)
        for gfid in sys.stdin:
            gfid_to_path(gfid)


if __name__ == "__main__":
    main()



#
# Copyright (c) 2016 Red Hat, Inc. <http://www.redhat.com>
# This file is part of GlusterFS.

# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.
#

GLUSTERFS_LIBEXECDIR = '/usr/local/libexec/glusterfs'
GLUSTERD_WORKDIR = "/var/lib/glusterd"

LOCALSTATEDIR = "/var"
UUID_FILE = "/var/lib/glusterd/glusterd.info"
GLUSTERFS_CONFDIR = "/usr/local/etc/glusterfs"
GCONF_VERSION = 4.0




import os
import sys
from optparse import OptionParser

import xattr

def convert(string):
    tmp_string = string
    if (string[0] == '0' and
        (string[1] == 's' or
         string[1] == 'S')):
        tmp_string = string.strip('%s%s' %
                                  (string[0],
                                   string[1]))
        return tmp_string.decode('base64')

    if (string[0] == '0' and
        (string[1] == 'x' or
         string[1] == 'X')):
        tmp_string = string.split('%s%s' %
                                  (string[0],
                                   string[1]))
        return tmp_string[1].decode('hex')

    return tmp_string

if __name__ == '__main__':
    usage = "usage: %prog [-n name] [-v value] [-x name]"
    parser = OptionParser(usage=usage)
    parser.add_option("-n", action="store", dest="name", type="string",
                      help="Specifies the name of the extended attribute to set.")
    parser.add_option("-v", action="store", dest="value", type="string",
                      help="Specifies the new value of the extended attribute."
                      " There are three methods available for encoding the value."
                      " If the given string is enclosed in double quotes, the"
                      " inner string is treated as text. In that case,"
                      " backslashes and double quotes have special meanings"
                      " and need to be escaped by a preceding backslash. Any"
                      " control characters can be encoded as a backslash"
                      " followed by three digits as its ASCII code in octal."
                      " If the given string begins with 0x or 0X, it expresses"
                      " a hexadecimal number. If the given string begins with"
                      " 0s or 0S, base64 encoding is expected.")
    parser.add_option("-x", action="store", dest="xname", type="string",
                      help="Remove the named extended attribute entirely.")

    (option, args) = parser.parse_args()
    if not args:
        print ("Usage: setfattr {-n name} [-v value] file...")
        print ("       setfattr {-x name} file...")
        print ("Try `setfattr --help' for more information.")
        sys.exit(1)

    if option.name and option.xname:
        print ("-n and -x are mutually exclusive...")
        sys.exit(1)

    if option.name:
        if option.value is None:
            print ("-n option requires -v value...")

    args[0] = os.path.abspath(args[0])

    if option.name and option.value:
        try:
            xattr.setxattr(args[0], option.name, convert(option.value))
        except Exception as err:
            print (err)
            sys.exit(1)

    if option.xname:
        try:
            xattr.removexattr(args[0], option.xname)
        except Exception as err:
            print (err)
            sys.exit(1)



#
# Copyright (c) 2011-2014 Red Hat, Inc. <http://www.redhat.com>
# This file is part of GlusterFS.

# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.
#

import re
import os
import sys
import stat
import time
import fcntl
import types
import struct
import logging
import tempfile
import subprocess
from errno import (EEXIST, ENOENT, ENODATA, ENOTDIR, ELOOP, EACCES,
                   EISDIR, ENOTEMPTY, ESTALE, EINVAL, EBUSY, EPERM)
import errno

from rconf import rconf
import gsyncdconfig as gconf
import libgfchangelog

import repce
from repce import RepceServer, RepceClient
from master import gmaster_builder
import syncdutils
from syncdutils import (GsyncdError, select, privileged, funcode,
                        entry2pb, gauxpfx, errno_wrap, lstat,
                        NoStimeAvailable, PartialHistoryAvailable,
                        ChangelogException, ChangelogHistoryNotAvailable,
                        get_changelog_log_level, get_rsync_version,
                        GX_GFID_CANONICAL_LEN,
                        gf_mount_ready, lf, Popen, sup,
                        Xattr, matching_disk_gfid, get_gfid_from_mnt,
                        unshare_propagation_supported, get_slv_dir_path)
from gsyncdstatus import GeorepStatus
from py2py3 import (pipe, str_to_bytearray, entry_pack_reg,
                    entry_pack_reg_stat, entry_pack_mkdir,
                    entry_pack_symlink)


ENOTSUP = getattr(errno, 'ENOTSUP', 'EOPNOTSUPP')

slv_volume = None
slv_host = None


class Server(object):

    """singleton implemening those filesystem access primitives
       which are needed for geo-replication functionality

    (Singleton in the sense it's a class which has only static
    and classmethods and is used directly, without instantiation.)
    """

    GX_NSPACE_PFX = (privileged() and "trusted" or "system")
    GX_NSPACE = GX_NSPACE_PFX + ".glusterfs"
    NTV_FMTSTR = "!" + "B" * 19 + "II"
    FRGN_XTRA_FMT = "I"
    FRGN_FMTSTR = NTV_FMTSTR + FRGN_XTRA_FMT

    # for backend gfid fetch, do not use GX_NSPACE_PFX
    GFID_XATTR = 'trusted.gfid'
    GFID_FMTSTR = "!" + "B" * 16

    local_path = ''

    @classmethod
    def _fmt_mknod(cls, l):
        return "!II%dsI%dsIII" % (GX_GFID_CANONICAL_LEN, l + 1)

    @classmethod
    def _fmt_mkdir(cls, l):
        return "!II%dsI%dsII" % (GX_GFID_CANONICAL_LEN, l + 1)

    @classmethod
    def _fmt_symlink(cls, l1, l2):
        return "!II%dsI%ds%ds" % (GX_GFID_CANONICAL_LEN, l1 + 1, l2 + 1)

    def _pathguard(f):
        """decorator method that checks
        the path argument of the decorated
        functions to make sure it does not
        point out of the managed tree
        """

        fc = funcode(f)
        pi = list(fc.co_varnames).index('path')

        def ff(*args):
            path = args[pi]
            ps = path.split('/')
            if path[0] == '/' or '..' in ps:
                raise ValueError('unsafe path')
            args = list(args)
            args[pi] = os.path.join(args[0].local_path, path)
            return f(*args)
        return ff

    @classmethod
    @_pathguard
    def entries(cls, path):
        """directory entries in an array"""
        # prevent symlinks being followed
        if not stat.S_ISDIR(os.lstat(path).st_mode):
            raise OSError(ENOTDIR, os.strerror(ENOTDIR))
        return os.listdir(path)

    @classmethod
    @_pathguard
    def lstat(cls, path):
        try:
            return os.lstat(path)
        except (IOError, OSError):
            ex = sys.exc_info()[1]
            if ex.errno == ENOENT:
                return ex.errno
            else:
                raise

    @classmethod
    @_pathguard
    def linkto_check(cls, path):
        try:
            return not (
                Xattr.lgetxattr_buf(path,
                                    'trusted.glusterfs.dht.linkto') == '')
        except (IOError, OSError):
            ex = sys.exc_info()[1]
            if ex.errno in (ENOENT, ENODATA):
                return False
            else:
                raise

    @classmethod
    @_pathguard
    def gfid(cls, path):
        buf = errno_wrap(Xattr.lgetxattr, [path, cls.GFID_XATTR, 16],
                         [ENOENT], [ESTALE, ENODATA])
        if buf == ENOENT:
            return buf
        else:
            buf = str_to_bytearray(buf)
            m = re.match('(.{8})(.{4})(.{4})(.{4})(.{12})', "".join(
                ['%02x' % x for x in struct.unpack(cls.GFID_FMTSTR, buf)]))
            return '-'.join(m.groups())

    @classmethod
    @_pathguard
    def purge(cls, path, entries=None):
        """force-delete subtrees

        If @entries is not specified, delete
        the whole subtree under @path (including
        @path).

        Otherwise, @entries should be a
        a sequence of children of @path, and
        the effect is identical with a joint
        @entries-less purge on them, ie.

        for e in entries:
            cls.purge(os.path.join(path, e))
        """
        me_also = entries is None
        if not entries:
            try:
                # if it's a symlink, prevent
                # following it
                try:
                    os.unlink(path)
                    return
                except OSError:
                    ex = sys.exc_info()[1]
                    if ex.errno == EISDIR:
                        entries = os.listdir(path)
                    else:
                        raise
            except OSError:
                ex = sys.exc_info()[1]
                if ex.errno in (ENOTDIR, ENOENT, ELOOP):
                    try:
                        os.unlink(path)
                        return
                    except OSError:
                        ex = sys.exc_info()[1]
                        if ex.errno == ENOENT:
                            return
                        raise
                else:
                    raise
        for e in entries:
            cls.purge(os.path.join(path, e))
        if me_also:
            os.rmdir(path)

    @classmethod
    @_pathguard
    def _create(cls, path, ctor):
        """path creation backend routine"""
        try:
            ctor(path)
        except OSError:
            ex = sys.exc_info()[1]
            if ex.errno == EEXIST:
                cls.purge(path)
                return ctor(path)
            raise

    @classmethod
    @_pathguard
    def mkdir(cls, path):
        cls._create(path, os.mkdir)

    @classmethod
    @_pathguard
    def symlink(cls, lnk, path):
        cls._create(path, lambda p: os.symlink(lnk, p))

    @classmethod
    @_pathguard
    def xtime(cls, path, uuid):
        """query xtime extended attribute

        Return xtime of @path for @uuid as a pair of integers.
        "Normal" errors due to non-existent @path or extended attribute
        are tolerated and errno is returned in such a case.
        """

        try:
            val = Xattr.lgetxattr(path,
                                  '.'.join([cls.GX_NSPACE, uuid, 'xtime']),
                                  8)
            val = str_to_bytearray(val)
            return struct.unpack('!II', val)
        except OSError:
            ex = sys.exc_info()[1]
            if ex.errno in (ENOENT, ENODATA, ENOTDIR):
                return ex.errno
            else:
                raise

    @classmethod
    @_pathguard
    def stime_mnt(cls, path, uuid):
        """query xtime extended attribute

        Return xtime of @path for @uuid as a pair of integers.
        "Normal" errors due to non-existent @path or extended attribute
        are tolerated and errno is returned in such a case.
        """

        try:
            val = Xattr.lgetxattr(path,
                                  '.'.join([cls.GX_NSPACE, uuid, 'stime']),
                                  8)
            val = str_to_bytearray(val)
            return struct.unpack('!II', val)
        except OSError:
            ex = sys.exc_info()[1]
            if ex.errno in (ENOENT, ENODATA, ENOTDIR):
                return ex.errno
            else:
                raise

    @classmethod
    @_pathguard
    def stime(cls, path, uuid):
        """query xtime extended attribute

        Return xtime of @path for @uuid as a pair of integers.
        "Normal" errors due to non-existent @path or extended attribute
        are tolerated and errno is returned in such a case.
        """

        try:
            val = Xattr.lgetxattr(path,
                                  '.'.join([cls.GX_NSPACE, uuid, 'stime']),
                                  8)
            val = str_to_bytearray(val)
            return struct.unpack('!II', val)
        except OSError:
            ex = sys.exc_info()[1]
            if ex.errno in (ENOENT, ENODATA, ENOTDIR):
                return ex.errno
            else:
                raise

    @classmethod
    @_pathguard
    def entry_stime(cls, path, uuid):
        """
        entry_stime xattr to reduce the number of retry of Entry changes when
        Geo-rep worker crashes and restarts. entry_stime is updated after
        processing every changelog file. On failure and restart, worker only
        have to reprocess the last changelog for Entry ops.
        Xattr Key: <PFX>.<MASTERVOL_UUID>.<SLAVEVOL_UUID>.entry_stime
        """
        try:
            val = Xattr.lgetxattr(path,
                                  '.'.join([cls.GX_NSPACE, uuid,
                                            'entry_stime']),
                                  8)
            val = str_to_bytearray(val)
            return struct.unpack('!II', val)
        except OSError:
            ex = sys.exc_info()[1]
            if ex.errno in (ENOENT, ENODATA, ENOTDIR):
                return ex.errno
            else:
                raise

    @classmethod
    def node_uuid(cls, path='.'):
        try:
            uuid_l = Xattr.lgetxattr_buf(
                path, '.'.join([cls.GX_NSPACE, 'node-uuid']))
            return uuid_l[:-1].split(' ')
        except OSError:
            raise

    @classmethod
    @_pathguard
    def set_stime(cls, path, uuid, mark):
        """set @mark as stime for @uuid on @path"""
        errno_wrap(Xattr.lsetxattr,
                   [path,
                    '.'.join([cls.GX_NSPACE, uuid, 'stime']),
                    struct.pack('!II', *mark)],
                   [ENOENT],
                   [ESTALE, EINVAL])

    @classmethod
    @_pathguard
    def set_entry_stime(cls, path, uuid, mark):
        """set @mark as stime for @uuid on @path"""
        errno_wrap(Xattr.lsetxattr,
                   [path,
                    '.'.join([cls.GX_NSPACE, uuid, 'entry_stime']),
                    struct.pack('!II', *mark)],
                   [ENOENT],
                   [ESTALE, EINVAL])

    @classmethod
    @_pathguard
    def set_xtime(cls, path, uuid, mark):
        """set @mark as xtime for @uuid on @path"""
        errno_wrap(Xattr.lsetxattr,
                   [path,
                    '.'.join([cls.GX_NSPACE, uuid, 'xtime']),
                    struct.pack('!II', *mark)],
                   [ENOENT],
                   [ESTALE, EINVAL])

    @classmethod
    @_pathguard
    def set_xtime_remote(cls, path, uuid, mark):
        """
        set @mark as xtime for @uuid on @path
        the difference b/w this and set_xtime() being
        set_xtime() being overloaded to set the xtime
        on the brick (this method sets xtime on the
        remote slave)
        """
        Xattr.lsetxattr(
            path, '.'.join([cls.GX_NSPACE, uuid, 'xtime']),
            struct.pack('!II', *mark))

    @classmethod
    def entry_ops(cls, entries):
        pfx = gauxpfx()
        logging.debug('entries: %s' % repr(entries))
        dist_count = rconf.args.master_dist_count

        def entry_purge(op, entry, gfid, e, uid, gid):
            # This is an extremely racy code and needs to be fixed ASAP.
            # The GFID check here is to be sure that the pargfid/bname
            # to be purged is the GFID gotten from the changelog.
            # (a stat(changelog_gfid) would also be valid here)
            # The race here is between the GFID check and the purge.

            # If the entry or the gfid of the file to be deleted is not present
            # on slave, we can ignore the unlink/rmdir
            if isinstance(lstat(entry), int) or \
               isinstance(lstat(os.path.join(pfx, gfid)), int):
                return

            if not matching_disk_gfid(gfid, entry):
                collect_failure(e, EEXIST, uid, gid)
                return

            if op == 'UNLINK':
                er = errno_wrap(os.unlink, [entry], [ENOENT, ESTALE], [EBUSY])
                # EISDIR is safe error, ignore. This can only happen when
                # unlink is sent from master while fixing gfid conflicts.
                if er != EISDIR:
                    return er

            elif op == 'RMDIR':
                er = errno_wrap(os.rmdir, [entry], [ENOENT, ESTALE,
                                                    ENOTEMPTY], [EBUSY])
                if er == ENOTEMPTY:
                    return er

        def collect_failure(e, cmd_ret, uid, gid, dst=False):
            slv_entry_info = {}
            slv_entry_info['gfid_mismatch'] = False
            slv_entry_info['name_mismatch'] = False
            slv_entry_info['dst'] = dst
            slv_entry_info['slave_isdir'] = False
            slv_entry_info['slave_name'] = None
            slv_entry_info['slave_gfid'] = None
            # We do this for failing fops on Slave
            # Master should be logging this
            if cmd_ret is None:
                return False

            if e.get("stat", {}):
                # Copy actual UID/GID value back to entry stat
                e['stat']['uid'] = uid
                e['stat']['gid'] = gid

            if cmd_ret in [EEXIST, ESTALE]:
                if dst:
                    en = e['entry1']
                else:
                    en = e['entry']
                disk_gfid = get_gfid_from_mnt(en)
                if isinstance(disk_gfid, str) and \
                   e['gfid'] != disk_gfid:
                    slv_entry_info['gfid_mismatch'] = True
                    st = lstat(en)
                    if not isinstance(st, int):
                        if st and stat.S_ISDIR(st.st_mode):
                            slv_entry_info['slave_isdir'] = True
                            dir_name = get_slv_dir_path(slv_host, slv_volume,
                                                        disk_gfid)
                            slv_entry_info['slave_name'] = dir_name
                        else:
                            slv_entry_info['slave_isdir'] = False
                    slv_entry_info['slave_gfid'] = disk_gfid
                    failures.append((e, cmd_ret, slv_entry_info))
                else:
                    return False
            else:
                failures.append((e, cmd_ret, slv_entry_info))

            return True

        failures = []

        def recursive_rmdir(gfid, entry, path):
            """disk_gfid check added for original path for which
            recursive_delete is called. This disk gfid check executed
            before every Unlink/Rmdir. If disk gfid is not matching
            with GFID from Changelog, that means other worker
            deleted the directory. Even if the subdir/file present,
            it belongs to different parent. Exit without performing
            further deletes.
            """
            if not matching_disk_gfid(gfid, entry):
                return

            names = []
            names = errno_wrap(os.listdir, [path], [ENOENT], [ESTALE, ENOTSUP])
            if isinstance(names, int):
                return

            for name in names:
                fullname = os.path.join(path, name)
                if not matching_disk_gfid(gfid, entry):
                    return
                er = errno_wrap(os.remove, [fullname], [ENOENT, ESTALE,
                                                        EISDIR], [EBUSY])

                if er == EISDIR:
                    recursive_rmdir(gfid, entry, fullname)

            if not matching_disk_gfid(gfid, entry):
                return

            errno_wrap(os.rmdir, [path], [ENOENT, ESTALE], [EBUSY])

        def rename_with_disk_gfid_confirmation(gfid, entry, en, uid, gid):
            if not matching_disk_gfid(gfid, entry):
                logging.error(lf("RENAME ignored: source entry does not match "
                                 "with on-disk gfid",
                                 source=entry,
                                 gfid=gfid,
                                 disk_gfid=get_gfid_from_mnt(entry),
                                 target=en))
                collect_failure(e, EEXIST, uid, gid)
                return

            cmd_ret = errno_wrap(os.rename,
                                 [entry, en],
                                 [ENOENT, EEXIST], [ESTALE, EBUSY])
            collect_failure(e, cmd_ret, uid, gid)

        for e in entries:
            blob = None
            op = e['op']
            gfid = e['gfid']
            entry = e['entry']
            uid = 0
            gid = 0

            # Skip entry processing if it's marked true during gfid
            # conflict resolution
            if e['skip_entry']:
                continue

            if e.get("stat", {}):
                # Copy UID/GID value and then reset to zero. Copied UID/GID
                # will be used to run chown once entry is created.
                uid = e['stat']['uid']
                gid = e['stat']['gid']
                e['stat']['uid'] = 0
                e['stat']['gid'] = 0

            (pg, bname) = entry2pb(entry)
            if op in ['RMDIR', 'UNLINK']:
                # Try once, if rmdir failed with ENOTEMPTY
                # then delete recursively.
                er = entry_purge(op, entry, gfid, e, uid, gid)
                if isinstance(er, int):
                    if er == ENOTEMPTY and op == 'RMDIR':
                        # Retry if ENOTEMPTY, ESTALE
                        er1 = errno_wrap(recursive_rmdir,
                                         [gfid, entry,
                                          os.path.join(pg, bname)],
                                         [], [ENOTEMPTY, ESTALE, ENODATA])
                        if not isinstance(er1, int):
                            logging.debug("Removed %s => %s/%s recursively" %
                                          (gfid, pg, bname))
                        else:
                            logging.warn(lf("Recursive remove failed",
                                            gfid=gfid,
                                            pgfid=pg,
                                            bname=bname,
                                            error=os.strerror(er1)))
                    else:
                        logging.warn(lf("Failed to remove",
                                        gfid=gfid,
                                        pgfid=pg,
                                        bname=bname,
                                        error=os.strerror(er)))
            elif op in ['CREATE', 'MKNOD']:
                slink = os.path.join(pfx, gfid)
                st = lstat(slink)
                # don't create multiple entries with same gfid
                if isinstance(st, int):
                    blob = entry_pack_reg(cls, gfid, bname,
                                          e['mode'], e['uid'], e['gid'])
                # Self healed hardlinks are recorded as MKNOD.
                # So if the gfid already exists, it should be
                # processed as hard link not mknod.
                elif op in ['MKNOD']:
                    cmd_ret = errno_wrap(os.link,
                                         [slink, entry],
                                         [ENOENT, EEXIST], [ESTALE])
                    collect_failure(e, cmd_ret, uid, gid)
            elif op == 'MKDIR':
                en = e['entry']
                slink = os.path.join(pfx, gfid)
                st = lstat(slink)
                # don't create multiple entries with same gfid
                if isinstance(st, int):
                    blob = entry_pack_mkdir(cls, gfid, bname,
                                            e['mode'], e['uid'], e['gid'])
                elif (isinstance(lstat(en), int) or
                      not matching_disk_gfid(gfid, en)):
                    # If gfid of a directory exists on slave but path based
                    # create is getting EEXIST. This means the directory is
                    # renamed in master but recorded as MKDIR during hybrid
                    # crawl. Get the directory path by reading the backend
                    # symlink and trying to rename to new name as said by
                    # master.
                    logging.info(lf("Special case: rename on mkdir",
                                    gfid=gfid, entry=repr(entry)))
                    src_entry = get_slv_dir_path(slv_host, slv_volume, gfid)
                    if src_entry is None:
                        collect_failure(e, ENOENT, uid, gid)
                    if src_entry is not None and src_entry != entry:
                        slv_entry_info = {}
                        slv_entry_info['gfid_mismatch'] = False
                        slv_entry_info['name_mismatch'] = True
                        slv_entry_info['dst'] = False
                        slv_entry_info['slave_isdir'] = True
                        slv_entry_info['slave_gfid'] = gfid
                        slv_entry_info['slave_entry'] = src_entry

                        failures.append((e, EEXIST, slv_entry_info))
            elif op == 'LINK':
                slink = os.path.join(pfx, gfid)
                st = lstat(slink)
                if isinstance(st, int):
                    (pg, bname) = entry2pb(entry)
                    if stat.S_ISREG(e['stat']['mode']):
                        blob = entry_pack_reg_stat(cls, gfid, bname, e['stat'])
                    elif stat.S_ISLNK(e['stat']['mode']):
                        blob = entry_pack_symlink(cls, gfid, bname, e['link'],
                                                  e['stat'])
                else:
                    cmd_ret = errno_wrap(os.link,
                                         [slink, entry],
                                         [ENOENT, EEXIST], [ESTALE])
                    collect_failure(e, cmd_ret, uid, gid)
            elif op == 'SYMLINK':
                en = e['entry']
                st = lstat(entry)
                if isinstance(st, int):
                    blob = entry_pack_symlink(cls, gfid, bname, e['link'],
                                              e['stat'])
                elif not matching_disk_gfid(gfid, en):
                    collect_failure(e, EEXIST, uid, gid)
            elif op == 'RENAME':
                en = e['entry1']
                # The matching disk gfid check validates two things
                #  1. Validates name is present, return false otherwise
                #  2. Validates gfid is same, returns false otherwise
                # So both validations are necessary to decide src doesn't
                # exist. We can't rely on only gfid stat as hardlink could
                # be present and we can't rely only on name as name could
                # exist with different gfid.
                if not matching_disk_gfid(gfid, entry):
                    if e['stat'] and not stat.S_ISDIR(e['stat']['mode']):
                        if stat.S_ISLNK(e['stat']['mode']):
                            # src is not present, so don't sync symlink as
                            # we don't know target. It's ok to ignore. If
                            # it's unliked, it's fine. If it's renamed to
                            # something else, it will be synced then.
                            if e['link'] is not None:
                                st1 = lstat(en)
                                if isinstance(st1, int):
                                    (pg, bname) = entry2pb(en)
                                    blob = entry_pack_symlink(cls, gfid, bname,
                                                              e['link'],
                                                              e['stat'])
                                elif not matching_disk_gfid(gfid, en):
                                    collect_failure(e, EEXIST, uid, gid, True)
                        else:
                            slink = os.path.join(pfx, gfid)
                            st = lstat(slink)
                            # don't create multiple entries with same gfid
                            if isinstance(st, int):
                                (pg, bname) = entry2pb(en)
                                blob = entry_pack_reg_stat(cls, gfid, bname,
                                                           e['stat'])
                            else:
                                cmd_ret = errno_wrap(os.link, [slink, en],
                                                    [ENOENT, EEXIST], [ESTALE])
                                collect_failure(e, cmd_ret, uid, gid)
                else:
                    st = lstat(entry)
                    st1 = lstat(en)
                    if isinstance(st1, int):
                        rename_with_disk_gfid_confirmation(gfid, entry, en,
                                                           uid, gid)
                    else:
                        if st.st_ino == st1.st_ino:
                            # we have a hard link, we can now unlink source
                            try:
                                errno_wrap(os.unlink, [entry],
                                           [ENOENT, ESTALE], [EBUSY])
                            except OSError as e:
                                if e.errno == EISDIR:
                                    try:
                                        errno_wrap(os.rmdir, [entry],
                                                   [ENOENT, ESTALE], [EBUSY])
                                    except OSError as e:
                                        if e.errno == ENOTEMPTY:
                                            logging.error(
                                                lf("Directory Rename failed. "
                                                   "Both Old and New"
                                                   " directories exists",
                                                   old=entry,
                                                   new=en))
                                        else:
                                            raise
                                else:
                                    raise
                        elif not matching_disk_gfid(gfid, en) and dist_count > 1:
                            collect_failure(e, EEXIST, uid, gid, True)
                        else:
                            # We are here which means matching_disk_gfid for
                            # both source and destination has returned false
                            # and distribution count for master vol is greater
                            # then one. Which basically says both the source and
                            # destination exist and not hardlinks.
                            # So we are safe to go ahead with rename here.
                            rename_with_disk_gfid_confirmation(gfid, entry, en,
                                                               uid, gid)
            if blob:
                cmd_ret = errno_wrap(Xattr.lsetxattr,
                                     [pg, 'glusterfs.gfid.newfile', blob],
                                     [EEXIST, ENOENT, ESTALE],
                                     [ESTALE, EINVAL, EBUSY])
                collect_failure(e, cmd_ret, uid, gid)

                # If UID/GID is different than zero that means we are trying
                # create Entry with different UID/GID. Create Entry with
                # UID:0 and GID:0, and then call chown to set UID/GID
                if uid != 0 or gid != 0:
                    path = os.path.join(pfx, gfid)
                    cmd_ret = errno_wrap(os.lchown, [path, uid, gid], [ENOENT],
                                         [ESTALE, EINVAL])
                    collect_failure(e, cmd_ret, uid, gid)

        return failures

    @classmethod
    def meta_ops(cls, meta_entries):
        logging.debug('Meta-entries: %s' % repr(meta_entries))
        failures = []
        for e in meta_entries:
            mode = e['stat']['mode']
            uid = e['stat']['uid']
            gid = e['stat']['gid']
            atime = e['stat']['atime']
            mtime = e['stat']['mtime']
            go = e['go']
            # Linux doesn't support chmod on symlink itself.
            # It is always applied to the target file. So
            # changelog would record target file's gfid
            # and we are good. But 'chown' is supported on
            # symlink file. So changelog would record symlink
            # gfid in such cases. Since we do 'chown' 'chmod'
            # 'utime' for each gfid recorded for metadata, and
            # we know from changelog the metadata is on symlink's
            # gfid or target file's gfid, we should be doing
            # 'lchown' 'lchmod' 'utime with no-deference' blindly.
            # But since 'lchmod' and 'utime with no de-reference' is
            # not supported in python3, we have to rely on 'chmod'
            # and 'utime with de-reference'. Hence avoiding 'chmod'
            # and 'utime' if it's symlink file.

            is_symlink = False
            cmd_ret = errno_wrap(os.lchown, [go, uid, gid], [ENOENT],
                                 [ESTALE, EINVAL])
            if isinstance(cmd_ret, int):
                continue

            is_symlink = os.path.islink(go)

            if not is_symlink:
                cmd_ret = errno_wrap(os.chmod, [go, mode],
                                     [ENOENT, EACCES, EPERM], [ESTALE, EINVAL])
                if isinstance(cmd_ret, int):
                    failures.append((e, cmd_ret, "chmod"))

                cmd_ret = errno_wrap(os.utime, [go, (atime, mtime)],
                                     [ENOENT, EACCES, EPERM], [ESTALE, EINVAL])
                if isinstance(cmd_ret, int):
                    failures.append((e, cmd_ret, "utime"))
        return failures

    @classmethod
    @_pathguard
    def setattr(cls, path, adct):
        """set file attributes

        @adct is a dict, where 'own', 'mode' and 'times'
        keys are looked for and values used to perform
        chown, chmod or utimes on @path.
        """
        own = adct.get('own')
        if own:
            os.lchown(path, *own)
        mode = adct.get('mode')
        if mode:
            os.chmod(path, stat.S_IMODE(mode))
        times = adct.get('times')
        if times:
            os.utime(path, times)

    @staticmethod
    def pid():
        return os.getpid()

    last_keep_alive = 0

    @classmethod
    def keep_alive(cls, dct):
        """process keepalive messages.

        Return keep-alive counter (number of received keep-alive
        messages).

        Now the "keep-alive" message can also have a payload which is
        used to set a foreign volume-mark on the underlying file system.
        """
        if dct:
            key = '.'.join([cls.GX_NSPACE, 'volume-mark', dct['uuid']])
            val = struct.pack(cls.FRGN_FMTSTR,
                              *(dct['version'] +
                                tuple(int(x, 16)
                                      for x in re.findall('(?:[\da-f]){2}',
                                                          dct['uuid'])) +
                                (dct['retval'],) + dct['volume_mark'][0:2] + (
                                    dct['timeout'],)))
            Xattr.lsetxattr('.', key, val)
        cls.last_keep_alive += 1
        return cls.last_keep_alive

    @staticmethod
    def version():
        """version used in handshake"""
        return 1.0


class Mounter(object):

    """Abstract base class for mounter backends"""

    def __init__(self, params):
        self.params = params
        self.mntpt = None
        self.umount_cmd = []

    @classmethod
    def get_glusterprog(cls):
        gluster_cmd_dir = gconf.get("gluster-command-dir")
        if rconf.args.subcmd == "slave":
            gluster_cmd_dir = gconf.get("slave-gluster-command-dir")
        return os.path.join(gluster_cmd_dir, cls.glusterprog)

    def umount_l(self, d):
        """perform lazy umount"""
        po = Popen(self.make_umount_argv(d), stderr=subprocess.PIPE,
                   universal_newlines=True)
        po.wait()
        return po

    @classmethod
    def make_umount_argv(cls, d):
        raise NotImplementedError

    def make_mount_argv(self, label=None):
        raise NotImplementedError

    def cleanup_mntpt(self, *a):
        pass

    def handle_mounter(self, po):
        po.wait()

    def inhibit(self, label):
        """inhibit a gluster filesystem

        Mount glusterfs over a temporary mountpoint,
        change into the mount, and lazy unmount the
        filesystem.
        """
        mpi, mpo = pipe()
        mh = Popen.fork()
        if mh:
            # Parent
            os.close(mpi)
            fcntl.fcntl(mpo, fcntl.F_SETFD, fcntl.FD_CLOEXEC)
            d = None
            margv = self.make_mount_argv(label)
            if self.mntpt:
                # mntpt is determined pre-mount
                d = self.mntpt
                mnt_msg = d + '\0'
                encoded_msg = mnt_msg.encode()
                os.write(mpo, encoded_msg)
            po = Popen(margv, **self.mountkw)
            self.handle_mounter(po)
            po.terminate_geterr()
            logging.debug('auxiliary glusterfs mount in place')
            if not d:
                # mntpt is determined during mount
                d = self.mntpt
                mnt_msg = d + '\0'
                encoded_msg = mnt_msg.encode()
                os.write(mpo, encoded_msg)
            encoded_msg = 'M'.encode()
            os.write(mpo, encoded_msg)
            t = syncdutils.Thread(target=lambda: os.chdir(d))
            t.start()
            tlim = rconf.starttime + gconf.get("connection-timeout")
            while True:
                if not t.isAlive():
                    break

                if time.time() >= tlim:
                    syncdutils.finalize(exval=1)
                time.sleep(1)
            os.close(mpo)
            _, rv = syncdutils.waitpid(mh, 0)
            if rv:
                rv = (os.WIFEXITED(rv) and os.WEXITSTATUS(rv) or 0) - \
                     (os.WIFSIGNALED(rv) and os.WTERMSIG(rv) or 0)
                logging.warn(lf('stale mount possibly left behind',
                                path=d))
                raise GsyncdError("cleaning up temp mountpoint %s "
                                  "failed with status %d" %
                                  (d, rv))
        else:
            rv = 0
            try:
                os.setsid()
                os.close(mpo)
                mntdata = ''
                while True:
                    c = os.read(mpi, 1)
                    c = c.decode()
                    if not c:
                        break
                    mntdata += c
                if mntdata:
                    mounted = False
                    if mntdata[-1] == 'M':
                        mntdata = mntdata[:-1]
                        assert(mntdata)
                        mounted = True
                    assert(mntdata[-1] == '\0')
                    mntpt = mntdata[:-1]
                    assert(mntpt)

                    umount_master = False
                    umount_slave = False
                    if rconf.args.subcmd == "worker" \
                       and not unshare_propagation_supported() \
                       and not gconf.get("access-mount"):
                        umount_master = True
                    if rconf.args.subcmd == "slave" \
                       and not gconf.get("slave-access-mount"):
                        umount_slave = True

                    if mounted and (umount_master or umount_slave):
                        po = self.umount_l(mntpt)
                        po.terminate_geterr(fail_on_err=False)
                        if po.returncode != 0:
                            po.errlog()
                            rv = po.returncode
                        logging.debug("Lazy umount done: %s" % mntpt)
                    if umount_master or umount_slave:
                        self.cleanup_mntpt(mntpt)
            except:
                logging.exception('mount cleanup failure:')
                rv = 200
            os._exit(rv)

        #Polling the dht.subvol.status value.
        RETRIES = 10
        while not gf_mount_ready():
            if RETRIES < 0:
                logging.error('Subvols are not up')
                break
            RETRIES -= 1
            time.sleep(0.2)

        logging.debug('auxiliary glusterfs mount prepared')


class DirectMounter(Mounter):

    """mounter backend which calls mount(8), umount(8) directly"""

    mountkw = {'stderr': subprocess.PIPE, 'universal_newlines': True}
    glusterprog = 'glusterfs'

    @staticmethod
    def make_umount_argv(d):
        return ['umount', '-l', d]

    def make_mount_argv(self, label=None):
        self.mntpt = tempfile.mkdtemp(prefix='gsyncd-aux-mount-')
        rconf.mount_point = self.mntpt
        return [self.get_glusterprog()] + \
            ['--' + p for p in self.params] + [self.mntpt]

    def cleanup_mntpt(self, mntpt=None):
        if not mntpt:
            mntpt = self.mntpt
        errno_wrap(os.rmdir, [mntpt], [ENOENT, EBUSY])


class MountbrokerMounter(Mounter):

    """mounter backend using the mountbroker gluster service"""

    mountkw = {'stderr': subprocess.PIPE, 'stdout': subprocess.PIPE,
               'universal_newlines': True}
    glusterprog = 'gluster'

    @classmethod
    def make_cli_argv(cls):
        return [cls.get_glusterprog()] + ['--remote-host=localhost'] + \
            gconf.get("gluster-cli-options").split() + ['system::']

    @classmethod
    def make_umount_argv(cls, d):
        return cls.make_cli_argv() + ['umount', d, 'lazy']

    def make_mount_argv(self, label):
        return self.make_cli_argv() + \
            ['mount', label, 'user-map-root=' +
                syncdutils.getusername()] + self.params

    def handle_mounter(self, po):
        self.mntpt = po.stdout.readline()[:-1]
        rconf.mount_point = self.mntpt
        rconf.mountbroker = True
        self.umount_cmd = self.make_cli_argv() + ['umount']
        rconf.mbr_umount_cmd = self.umount_cmd
        po.stdout.close()
        sup(self, po)
        if po.returncode != 0:
            # if cli terminated with error due to being
            # refused by glusterd, what it put
            # out on stdout is a diagnostic message
            logging.error(lf('glusterd answered', mnt=self.mntpt))


class GLUSTERServer(Server):

    "server enhancements for a glusterfs backend"""

    @classmethod
    def _attr_unpack_dict(cls, xattr, extra_fields=''):
        """generic volume mark fetching/parsing backed"""
        fmt_string = cls.NTV_FMTSTR + extra_fields
        buf = Xattr.lgetxattr('.', xattr, struct.calcsize(fmt_string))
        buf = str_to_bytearray(buf)
        vm = struct.unpack(fmt_string, buf)
        m = re.match(
            '(.{8})(.{4})(.{4})(.{4})(.{12})',
            "".join(['%02x' % x for x in vm[2:18]]))
        uuid = '-'.join(m.groups())
        volinfo = {'version': vm[0:2],
                   'uuid': uuid,
                   'retval': vm[18],
                   'volume_mark': vm[19:21],
                   }
        if extra_fields:
            return volinfo, vm[-len(extra_fields):]
        else:
            return volinfo

    @classmethod
    def foreign_volume_infos(cls):
        """return list of valid (not expired) foreign volume marks"""
        dict_list = []
        xattr_list = Xattr.llistxattr_buf('.')
        for ele in xattr_list:
            if ele.find('.'.join([cls.GX_NSPACE, 'volume-mark', ''])) == 0:
                d, x = cls._attr_unpack_dict(ele, cls.FRGN_XTRA_FMT)
                now = int(time.time())
                if x[0] > now:
                    logging.debug("volinfo[%s] expires: %d "
                                  "(%d sec later)" %
                                  (d['uuid'], x[0], x[0] - now))
                    d['timeout'] = x[0]
                    dict_list.append(d)
                else:
                    try:
                        Xattr.lremovexattr('.', ele)
                    except OSError:
                        pass
        return dict_list

    @classmethod
    def native_volume_info(cls):
        """get the native volume mark of the underlying gluster volume"""
        try:
            return cls._attr_unpack_dict('.'.join([cls.GX_NSPACE,
                                                   'volume-mark']))
        except OSError:
            ex = sys.exc_info()[1]
            if ex.errno != ENODATA:
                raise


class GLUSTER(object):

    """scheme class for gluster:// urls

    can be used to represent a gluster slave server
    on slave side, or interface to a remote gluster
    slave on master side, or to represent master
    (slave-ish features come from the mixins, master
    functionality is outsourced to GMaster from master)
    """
    server = GLUSTERServer

    def __init__(self, host, volume):
        self.path = "%s:%s" % (host, volume)
        self.host = host
        self.volume = volume

        global slv_volume
        global slv_host
        slv_volume = self.volume
        slv_host = self.host

    def connect(self):
        """inhibit the resource beyond

        Choose mounting backend (direct or mountbroker),
        set up glusterfs parameters and perform the mount
        with given backend
        """

        logging.info("Mounting gluster volume locally...")
        t0 = time.time()
        label = gconf.get('mountbroker', None)
        if not label and not privileged():
            label = syncdutils.getusername()
        mounter = label and MountbrokerMounter or DirectMounter

        log_file = gconf.get("gluster-log-file")
        if rconf.args.subcmd == "slave":
            log_file = gconf.get("slave-gluster-log-file")

        log_level = gconf.get("gluster-log-level")
        if rconf.args.subcmd == "slave":
            log_level = gconf.get("slave-gluster-log-level")

        params = gconf.get("gluster-params").split() + \
            ['log-level=' + log_level] + \
            ['log-file=' + log_file, 'volfile-server=' + self.host] + \
            ['volfile-id=' + self.volume, 'client-pid=-1']

        self.mounter = mounter(params)
        self.mounter.inhibit(label)
        logging.info(lf("Mounted gluster volume",
                        duration="%.4f" % (time.time() - t0)))

    def gmaster_instantiate_tuple(self, slave):
        """return a tuple of the 'one shot' and the 'main crawl'
        class instance"""
        return (gmaster_builder('xsync')(self, slave),
                gmaster_builder()(self, slave),
                gmaster_builder('changeloghistory')(self, slave))

    def service_loop(self, slave=None):
        """enter service loop

        - if slave given, instantiate GMaster and
          pass control to that instance, which implements
          master behavior
        - else do that's what's inherited
        """
        if rconf.args.subcmd == "slave":
            if gconf.get("use-rsync-xattrs") and not privileged():
                raise GsyncdError(
                    "using rsync for extended attributes is not supported")

            repce = RepceServer(
                self.server, sys.stdin, sys.stdout, gconf.get("sync-jobs"))
            t = syncdutils.Thread(target=lambda: (repce.service_loop(),
                                                  syncdutils.finalize()))
            t.start()
            logging.info("slave listening")
            if gconf.get("slave-timeout") and gconf.get("slave-timeout") > 0:
                while True:
                    lp = self.server.last_keep_alive
                    time.sleep(gconf.get("slave-timeout"))
                    if lp == self.server.last_keep_alive:
                        logging.info(
                            lf("connection inactive, stopping",
                               timeout=gconf.get("slave-timeout")))
                        break
            else:
                select((), (), ())

            return

        class brickserver(Server):
            local_path = rconf.args.local_path
            aggregated = self.server

            @classmethod
            def entries(cls, path):
                e = super(brickserver, cls).entries(path)
                # on the brick don't mess with /.glusterfs
                if path == '.':
                    try:
                        e.remove('.glusterfs')
                        e.remove('.trashcan')
                    except ValueError:
                        pass
                return e

            @classmethod
            def lstat(cls, e):
                """ path based backend stat """
                return super(brickserver, cls).lstat(e)

            @classmethod
            def gfid(cls, e):
                """ path based backend gfid fetch """
                return super(brickserver, cls).gfid(e)

            @classmethod
            def linkto_check(cls, e):
                return super(brickserver, cls).linkto_check(e)

        # define {,set_}xtime in slave, thus preempting
        # the call to remote, so that it takes data from
        # the local brick
        slave.server.xtime = types.MethodType(
            lambda _self, path, uuid: (
                brickserver.xtime(path,
                                  uuid + '.' + rconf.args.slave_id)
            ),
            slave.server)
        slave.server.stime = types.MethodType(
            lambda _self, path, uuid: (
                brickserver.stime(path,
                                  uuid + '.' + rconf.args.slave_id)
            ),
            slave.server)
        slave.server.entry_stime = types.MethodType(
            lambda _self, path, uuid: (
                brickserver.entry_stime(
                    path,
                    uuid + '.' + rconf.args.slave_id)
            ),
            slave.server)
        slave.server.set_stime = types.MethodType(
            lambda _self, path, uuid, mark: (
                brickserver.set_stime(path,
                                      uuid + '.' + rconf.args.slave_id,
                                      mark)
            ),
            slave.server)
        slave.server.set_entry_stime = types.MethodType(
            lambda _self, path, uuid, mark: (
                brickserver.set_entry_stime(
                    path,
                    uuid + '.' + rconf.args.slave_id,
                    mark)
            ),
            slave.server)

        (g1, g2, g3) = self.gmaster_instantiate_tuple(slave)
        g1.master.server = brickserver
        g2.master.server = brickserver
        g3.master.server = brickserver

        # bad bad bad: bad way to do things like this
        # need to make this elegant
        # register the crawlers and start crawling
        # g1 ==> Xsync, g2 ==> config.change_detector(changelog by default)
        # g3 ==> changelog History
        status = GeorepStatus(gconf.get("state-file"),
                              rconf.args.local_node,
                              rconf.args.local_path,
                              rconf.args.local_node_id,
                              rconf.args.master,
                              rconf.args.slave)
        status.reset_on_worker_start()

        try:
            workdir = g2.setup_working_dir()
            # Register only when change_detector is not set to
            # xsync, else agent will generate changelog files
            # in .processing directory of working dir
            if gconf.get("change-detector") != 'xsync':
                # register with the changelog library
                # 9 == log level (DEBUG)
                # 5 == connection retries
                libgfchangelog.register(rconf.args.local_path,
                                        workdir,
                                        gconf.get("changelog-log-file"),
                                        get_changelog_log_level(
                                            gconf.get("changelog-log-level")),
                                        g2.CHANGELOG_CONN_RETRIES)

            register_time = int(time.time())
            g2.register(register_time, status)
            g3.register(register_time, status)
        except ChangelogException as e:
            logging.error(lf("Changelog register failed", error=e))
            sys.exit(1)

        g1.register(status=status)
        logging.info(lf("Register time",
                        time=register_time))
        # oneshot: Try to use changelog history api, if not
        # available switch to FS crawl
        # Note: if config.change_detector is xsync then
        # it will not use changelog history api
        try:
            g3.crawlwrap(oneshot=True)
        except PartialHistoryAvailable as e:
            logging.info(lf('Partial history available, using xsync crawl'
                            ' after consuming history',
                            till=e))
            g1.crawlwrap(oneshot=True, register_time=register_time)
        except ChangelogHistoryNotAvailable:
            logging.info('Changelog history not available, using xsync')
            g1.crawlwrap(oneshot=True, register_time=register_time)
        except NoStimeAvailable:
            logging.info('No stime available, using xsync crawl')
            g1.crawlwrap(oneshot=True, register_time=register_time)
        except ChangelogException as e:
            logging.error(lf("Changelog History Crawl failed",
                             error=e))
            sys.exit(1)

        try:
            g2.crawlwrap()
        except ChangelogException as e:
            logging.error(lf("Changelog crawl failed", error=e))
            sys.exit(1)


class SSH(object):

    """scheme class for ssh:// urls

    interface to remote slave on master side
    implementing an ssh based proxy
    """

    def __init__(self, host, volume):
        self.remote_addr = host
        self.volume = volume

    @staticmethod
    def parse_ssh_address(self):
        m = re.match('([^@]+)@(.+)', self.remote_addr)
        if m:
            u, h = m.groups()
        else:
            u, h = syncdutils.getusername(), self.remote_addr
        self.remotehost = h
        return {'user': u, 'host': h}

    def start_fd_client(self, i, o):
        """set up RePCe client, handshake with server

        It's cut out as a separate method to let
        subclasses hook into client startup
        """
        self.server = RepceClient(i, o)
        rv = self.server.__version__()
        exrv = {'proto': repce.repce_version, 'object': Server.version()}
        da0 = (rv, exrv)
        da1 = ({}, {})
        for i in range(2):
            for k, v in da0[i].items():
                da1[i][k] = int(v)
        if da1[0] != da1[1]:
            raise GsyncdError(
                "RePCe major version mismatch: local %s, remote %s" %
                (exrv, rv))
        slavepath = "/proc/%d/cwd" % self.server.pid()
        self.slaveurl = ':'.join([self.remote_addr, slavepath])

    def connect_remote(self):
        """connect to inner slave url through outer ssh url

        Wrap the connecting utility in ssh.

        Much care is put into daemonizing: in that case
        ssh is started before daemonization, but
        RePCe client is to be created after that (as ssh
        interactive password auth would be defeated by
        a daemonized ssh, while client should be present
        only in the final process). In that case the action
        is taken apart to two parts, this method is ivoked
        once pre-daemon, once post-daemon. Use @go_daemon
        to deiced what part to perform.

        [NB. ATM gluster product does not makes use of interactive
        authentication.]
        """
        syncdutils.setup_ssh_ctl(tempfile.mkdtemp(prefix='gsyncd-aux-ssh-'),
                                 self.remote_addr,
                                 self.volume)

        logging.info("Initializing SSH connection between master and slave...")
        t0 = time.time()

        extra_opts = []
        remote_gsyncd = gconf.get("remote-gsyncd")
        if remote_gsyncd == "":
            remote_gsyncd = "/nonexistent/gsyncd"

        if gconf.get("use-rsync-xattrs"):
            extra_opts.append('--use-rsync-xattrs')

        args_to_slave = [gconf.get("ssh-command")] + \
            gconf.get("ssh-options").split() + \
            ["-p", str(gconf.get("ssh-port"))] + \
            rconf.ssh_ctl_args + [self.remote_addr] + \
            [remote_gsyncd, "slave"] + \
            extra_opts + \
            [rconf.args.master, rconf.args.slave] + \
            [
                '--master-node', rconf.args.local_node,
                '--master-node-id', rconf.args.local_node_id,
                '--master-brick', rconf.args.local_path,
                '--local-node', rconf.args.resource_remote,
                '--local-node-id', rconf.args.resource_remote_id] + \
            [
                # Add all config arguments here, slave gsyncd will not use
                # config file in slave side, so all overriding options should
                # be sent as arguments
                '--slave-timeout', str(gconf.get("slave-timeout")),
                '--slave-log-level', gconf.get("slave-log-level"),
                '--slave-gluster-log-level',
                gconf.get("slave-gluster-log-level"),
                '--slave-gluster-command-dir',
                gconf.get("slave-gluster-command-dir"),
                '--master-dist-count',
                str(gconf.get("master-distribution-count"))]

        if gconf.get("slave-access-mount"):
            args_to_slave.append('--slave-access-mount')

        if rconf.args.debug:
            args_to_slave.append('--debug')

        po = Popen(args_to_slave,
                   stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                   stderr=subprocess.PIPE)
        rconf.transport = po
        self.start_fd_client(po.stdout, po.stdin)
        logging.info(lf("SSH connection between master and slave established.",
                        duration="%.4f" % (time.time() - t0)))

    def rsync(self, files, *args, **kw):
        """invoke rsync"""
        if not files:
            raise GsyncdError("no files to sync")
        logging.debug("files: " + ", ".join(files))

        extra_rsync_flags = []
        # Performance flag, --ignore-missing-args, if rsync version is
        # greater than 3.1.0 then include this flag.
        if gconf.get("rsync-opt-ignore-missing-args") and \
           get_rsync_version(gconf.get("rsync-command")) >= "3.1.0":
            extra_rsync_flags = ["--ignore-missing-args"]

        rsync_ssh_opts = [gconf.get("ssh-command")] + \
            gconf.get("ssh-options").split() + \
            ["-p", str(gconf.get("ssh-port"))] + \
            rconf.ssh_ctl_args + \
            gconf.get("rsync-ssh-options").split()

        argv = [
            gconf.get("rsync-command"),
            '-aR0',
            '--inplace',
            '--files-from=-',
            '--super',
            '--stats',
            '--numeric-ids',
            '--no-implied-dirs'
        ]

        if gconf.get("rsync-opt-existing"):
            argv += ["--existing"]

        if gconf.get("sync-xattrs"):
            argv += ['--xattrs']

        if gconf.get("sync-acls"):
            argv += ['--acls']

        argv = argv + \
            gconf.get("rsync-options").split() + \
            extra_rsync_flags + ['.'] + \
            ["-e", " ".join(rsync_ssh_opts)] + \
            [self.slaveurl]

        log_rsync_performance = gconf.getr("log-rsync-performance", False)

        rsync_verbose = False
        for item in argv:
            if item == "--verbose" or item.startswith("-v"):
                rsync_verbose = True
                break

        if log_rsync_performance or rsync_verbose:
            # use stdout=PIPE only when log_rsync_performance enabled
            # Else rsync will write to stdout and nobody is there
            # to consume. If PIPE is full rsync hangs.
            po = Popen(argv, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                       stderr=subprocess.PIPE, universal_newlines=True)
        else:
            po = Popen(argv, stdin=subprocess.PIPE, stderr=subprocess.PIPE,
                       universal_newlines=True)

        for f in files:
            po.stdin.write(f)
            po.stdin.write('\0')

        stdout, stderr = po.communicate()

        if kw.get("log_err", False):
            for errline in stderr.strip().split("\n")[:-1]:
                logging.error(lf("SYNC Error",
                                 sync_engine="Rsync",
                                 error=errline))

        if log_rsync_performance:
            rsync_msg = []
            for line in stdout.split("\n"):
                if line.startswith("Number of files:") or \
                   line.startswith("Number of regular files transferred:") or \
                   line.startswith("Total file size:") or \
                   line.startswith("Total transferred file size:") or \
                   line.startswith("Literal data:") or \
                   line.startswith("Matched data:") or \
                   line.startswith("Total bytes sent:") or \
                   line.startswith("Total bytes received:") or \
                   line.startswith("sent "):
                    rsync_msg.append(line)
            logging.info(lf("rsync performance",
                            data=", ".join(rsync_msg)))

        if rsync_verbose:
            logging.info(lf("rsync verbose", data=repr(stdout)))

        return po

    def tarssh(self, files, log_err=False):
        """invoke tar+ssh
        -z (compress) can be use if needed, but omitting it now
        as it results in weird error (tar+ssh errors out (errcode: 2)
        """
        if not files:
            raise GsyncdError("no files to sync")
        logging.debug("files: " + ", ".join(files))
        (host, rdir) = self.slaveurl.split(':')

        tar_cmd = ["tar"] + \
            ["--sparse", "-cf", "-", "--files-from", "-"]
        ssh_cmd = gconf.get("ssh-command").split() + \
            gconf.get("ssh-options-tar").split() + \
            ["-p", str(gconf.get("ssh-port"))] + \
            [host, "tar"] + \
            ["--overwrite", "-xf", "-", "-C", rdir]
        p0 = Popen(tar_cmd, stdout=subprocess.PIPE,
                   stdin=subprocess.PIPE, stderr=subprocess.PIPE,
                   universal_newlines=True)
        p1 = Popen(ssh_cmd, stdin=p0.stdout, stderr=subprocess.PIPE,
                   universal_newlines=True)
        for f in files:
            p0.stdin.write(f)
            p0.stdin.write('\n')

        p0.stdin.close()
        p0.stdout.close()  # Allow p0 to receive a SIGPIPE if p1 exits.

        # stdin and stdout of p0 is already closed, Reset to None and
        # wait for child process to complete
        p0.stdin = None
        p0.stdout = None

        def wait_for_tar(p0):
            _, stderr = p0.communicate()
            if log_err:
                for errline in stderr.strip().split("\n")[:-1]:
                    if "No such file or directory" not in errline:
                        logging.error(lf("SYNC Error",
                                         sync_engine="Tarssh",
                                         error=errline))

        t = syncdutils.Thread(target=wait_for_tar, args=(p0, ))
        # wait for tar to terminate, collecting any errors, further
        # waiting for transfer to complete
        t.start()

        # wait for ssh process
        _, stderr1 = p1.communicate()
        t.join()

        if log_err:
            for errline in stderr1.strip().split("\n")[:-1]:
                logging.error(lf("SYNC Error",
                                 sync_engine="Tarssh",
                                 error=errline))

        return p1



# -*- coding: utf-8 -*-

# Copyright (c) 2015 Red Hat, Inc. <http://www.redhat.com/>
# This file is part of GlusterFS.
#
# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.

import sys
from subprocess import PIPE, Popen
from errno import EEXIST, ENOENT
import xml.etree.cElementTree as etree
import logging
import os
from datetime import datetime

ROOT_GFID = "00000000-0000-0000-0000-000000000001"
DEFAULT_CHANGELOG_INTERVAL = 15
SPACE_ESCAPE_CHAR = "%20"
NEWLINE_ESCAPE_CHAR = "%0A"
PERCENTAGE_ESCAPE_CHAR = "%25"

ParseError = etree.ParseError if hasattr(etree, 'ParseError') else SyntaxError
cache_data = {}


class RecordType(object):
    NEW = "NEW"
    MODIFY = "MODIFY"
    RENAME = "RENAME"
    DELETE = "DELETE"


def cache_output(func):
    def wrapper(*args, **kwargs):
        global cache_data
        if cache_data.get(func.__name__, None) is None:
            cache_data[func.__name__] = func(*args, **kwargs)

        return cache_data[func.__name__]
    return wrapper


def handle_rm_error(func, path, exc_info):
    if exc_info[1].errno == ENOENT:
        return

    raise exc_info[1]


def find(path, callback_func=lambda x: True, filter_func=lambda x: True,
         ignore_dirs=[], subdirs_crawl=True):
    if path in ignore_dirs:
        return

    # Capture filter_func output and pass it to callback function
    filter_result = filter_func(path)
    if filter_result is not None:
        callback_func(path, filter_result, os.path.isdir(path))

    for p in os.listdir(path):
        full_path = os.path.join(path, p)

        is_dir = os.path.isdir(full_path)
        if is_dir:
            if subdirs_crawl:
                find(full_path, callback_func, filter_func, ignore_dirs)
            else:
                filter_result = filter_func(full_path)
                if filter_result is not None:
                    callback_func(full_path, filter_result)
        else:
            filter_result = filter_func(full_path)
            if filter_result is not None:
                callback_func(full_path, filter_result, is_dir)


def output_write(f, path, prefix=".", encode=False, tag="",
                 field_separator=" "):
    if path == "":
        return

    if prefix != ".":
        path = os.path.join(prefix, path)

    if encode:
        path = quote_plus_space_newline(path)

    # set the field separator
    FS = "" if tag == "" else field_separator

    f.write("%s%s%s\n" % (tag.strip(), FS, path))


def human_time(ts):
    return datetime.fromtimestamp(float(ts)).strftime("%Y-%m-%d %H:%M:%S")


def setup_logger(logger, path, debug=False):
    if debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    # create the logging file handler
    fh = logging.FileHandler(path)

    formatter = logging.Formatter("[%(asctime)s] %(levelname)s "
                                  "[%(module)s - %(lineno)s:%(funcName)s] "
                                  "- %(message)s")

    fh.setFormatter(formatter)

    # add handler to logger object
    logger.addHandler(fh)


def create_file(path, exit_on_err=False, logger=None):
    """
    If file exists overwrite. Print error to stderr and exit
    if exit_on_err is set, else raise the exception. Consumer
    should handle the exception.
    """
    try:
        open(path, 'w').close()
    except (OSError, IOError) as e:
        if exit_on_err:
            fail("Failed to create file %s: %s" % (path, e), logger=logger)
        else:
            raise


def mkdirp(path, exit_on_err=False, logger=None):
    """
    Try creating required directory structure
    ignore EEXIST and raise exception for rest of the errors.
    Print error in stderr and exit if exit_on_err is set, else
    raise exception.
    """
    try:
        os.makedirs(path)
    except (OSError, IOError) as e:
        if e.errno == EEXIST and os.path.isdir(path):
            pass
        else:
            if exit_on_err:
                fail("Fail to create dir %s: %s" % (path, e), logger=logger)
            else:
                raise


def fail(msg, code=1, logger=None):
    """
    Write error to stderr and exit
    """
    if logger:
        logger.error(msg)
    sys.stderr.write("%s\n" % msg)
    sys.exit(code)


def execute(cmd, exit_msg=None, logger=None):
    """
    If failure_msg is not None then return returncode, out and error.
    If failure msg is set, write to stderr and exit.
    """
    p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE, close_fds=True)

    (out, err) = p.communicate()
    if p.returncode != 0 and exit_msg is not None:
        fail("%s: %s" % (exit_msg, err), p.returncode, logger=logger)

    return (p.returncode, out, err)


def symlink_gfid_to_path(brick, gfid):
    """
    Each directories are symlinked to file named GFID
    in .glusterfs directory of brick backend. Using readlink
    we get PARGFID/basename of dir. readlink recursively till
    we get PARGFID as ROOT_GFID.
    """
    if gfid == ROOT_GFID:
        return ""

    out_path = ""
    while True:
        path = os.path.join(brick, ".glusterfs", gfid[0:2], gfid[2:4], gfid)
        path_readlink = os.readlink(path)
        pgfid = os.path.dirname(path_readlink)
        out_path = os.path.join(os.path.basename(path_readlink), out_path)
        if pgfid == "../../00/00/%s" % ROOT_GFID:
            break
        gfid = os.path.basename(pgfid)
    return out_path


@cache_output
def get_my_uuid():
    cmd = ["gluster", "system::", "uuid", "get", "--xml"]
    rc, out, err = execute(cmd)

    if rc != 0:
        return None

    tree = etree.fromstring(out)
    uuid_el = tree.find("uuidGenerate/uuid")
    return uuid_el.text


def is_host_local(host_uuid):
    # Get UUID only if it is not done previously
    # else Cache the UUID value
    my_uuid = get_my_uuid()
    if my_uuid == host_uuid:
        return True

    return False


def get_changelog_rollover_time(volumename):
    cmd = ["gluster", "volume", "get", volumename,
           "changelog.rollover-time", "--xml"]
    rc, out, err = execute(cmd)

    if rc != 0:
        return DEFAULT_CHANGELOG_INTERVAL

    try:
        tree = etree.fromstring(out)
        val = tree.find('volGetopts/Opt/Value').text
        if val is not None:
            # Filter the value by split, as it may be 'X (DEFAULT)'
            # and we only need 'X'
            return int(val.split(' ', 1)[0])
    except ParseError:
        return DEFAULT_CHANGELOG_INTERVAL


def output_path_prepare(path, args):
    """
    If Prefix is set, joins to Path, removes ending slash
    and encodes it.
    """
    if args.output_prefix != ".":
        path = os.path.join(args.output_prefix, path)
        if path.endswith("/"):
            path = path[0:len(path)-1]

    if args.no_encode:
        return path
    else:
        return quote_plus_space_newline(path)


def unquote_plus_space_newline(s):
    return s.replace(SPACE_ESCAPE_CHAR, " ")\
            .replace(NEWLINE_ESCAPE_CHAR, "\n")\
            .replace(PERCENTAGE_ESCAPE_CHAR, "%")


def quote_plus_space_newline(s):
    return s.replace("%", PERCENTAGE_ESCAPE_CHAR)\
            .replace(" ", SPACE_ESCAPE_CHAR)\
            .replace("\n", NEWLINE_ESCAPE_CHAR)



#
# Copyright (c) 2011-2014 Red Hat, Inc. <http://www.redhat.com>
# This file is part of GlusterFS.

# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.
#

import os
import sys
import time
import stat
import logging
import fcntl
import string
import errno
import tarfile
from errno import ENOENT, ENODATA, EEXIST, EACCES, EAGAIN, ESTALE, EINTR
from threading import Condition, Lock
from datetime import datetime

import gsyncdconfig as gconf
import libgfchangelog
from rconf import rconf
from syncdutils import (Thread, GsyncdError, escape_space_newline,
                        unescape_space_newline, gauxpfx, escape,
                        lstat, errno_wrap, FreeObject, lf, matching_disk_gfid,
                        NoStimeAvailable, PartialHistoryAvailable,
                        host_brick_split)

URXTIME = (-1, 0)

# Default rollover time set in changelog translator
# changelog rollover time is hardcoded here to avoid the
# xsync usage when crawling switch happens from history
# to changelog. If rollover time increased in translator
# then geo-rep can enter into xsync crawl after history
# crawl before starting live changelog crawl.
CHANGELOG_ROLLOVER_TIME = 15

# Utility functions to help us to get to closer proximity
# of the DRY principle (no, don't look for elevated or
# perspectivistic things here)


def _xtime_now():
    t = time.time()
    sec = int(t)
    nsec = int((t - sec) * 1000000)
    return (sec, nsec)


def _volinfo_hook_relax_foreign(self):
    volinfo_sys = self.get_sys_volinfo()
    fgn_vi = volinfo_sys[self.KFGN]
    if fgn_vi:
        expiry = fgn_vi['timeout'] - int(time.time()) + 1
        logging.info(lf('foreign volume info found, waiting for expiry',
                        expiry=expiry))
        time.sleep(expiry)
        volinfo_sys = self.get_sys_volinfo()
    return volinfo_sys


def edct(op, **ed):
    dct = {}
    dct['op'] = op
    # This is used in automatic gfid conflict resolution.
    # When marked True, it's skipped during re-processing.
    dct['skip_entry'] = False
    for k in ed:
        if k == 'stat':
            st = ed[k]
            dst = dct['stat'] = {}
            if st:
                dst['uid'] = st.st_uid
                dst['gid'] = st.st_gid
                dst['mode'] = st.st_mode
                dst['atime'] = st.st_atime
                dst['mtime'] = st.st_mtime
        else:
            dct[k] = ed[k]
    return dct


# The API!

def gmaster_builder(excrawl=None):
    """produce the GMaster class variant corresponding
       to sync mode"""
    this = sys.modules[__name__]
    modemixin = gconf.get("special-sync-mode")
    if not modemixin:
        modemixin = 'normal'

    if gconf.get("change-detector") == 'xsync':
        changemixin = 'xsync'
    elif excrawl:
        changemixin = excrawl
    else:
        changemixin = gconf.get("change-detector")

    logging.debug(lf('setting up change detection mode',
                     mode=changemixin))
    modemixin = getattr(this, modemixin.capitalize() + 'Mixin')
    crawlmixin = getattr(this, 'GMaster' + changemixin.capitalize() + 'Mixin')

    if gconf.get("use-rsync-xattrs"):
        sendmarkmixin = SendmarkRsyncMixin
    else:
        sendmarkmixin = SendmarkNormalMixin

    if gconf.get("ignore-deletes"):
        purgemixin = PurgeNoopMixin
    else:
        purgemixin = PurgeNormalMixin

    if gconf.get("sync-method") == "tarssh":
        syncengine = TarSSHEngine
    else:
        syncengine = RsyncEngine

    class _GMaster(crawlmixin, modemixin, sendmarkmixin,
                   purgemixin, syncengine):
        pass

    return _GMaster


# Mixin classes that implement the data format
# and logic particularities of the certain
# sync modes

class NormalMixin(object):

    """normal geo-rep behavior"""

    minus_infinity = URXTIME

    # following staticmethods ideally would be
    # methods of an xtime object (in particular,
    # implementing the hooks needed for comparison
    # operators), but at this point we don't yet
    # have a dedicated xtime class

    @staticmethod
    def serialize_xtime(xt):
        return "%d.%d" % tuple(xt)

    @staticmethod
    def deserialize_xtime(xt):
        return tuple(int(x) for x in xt.split("."))

    @staticmethod
    def native_xtime(xt):
        return xt

    @staticmethod
    def xtime_geq(xt0, xt1):
        return xt0 >= xt1

    def make_xtime_opts(self, is_master, opts):
        if 'create' not in opts:
            opts['create'] = is_master
        if 'default_xtime' not in opts:
            opts['default_xtime'] = URXTIME

    def xtime_low(self, rsc, path, **opts):
        if rsc == self.master:
            xt = rsc.server.xtime(path, self.uuid)
        else:
            xt = rsc.server.stime(path, self.uuid)
            if isinstance(xt, int) and xt == ENODATA:
                xt = rsc.server.xtime(path, self.uuid)
                if not isinstance(xt, int):
                    self.slave.server.set_stime(path, self.uuid, xt)
        if isinstance(xt, int) and xt != ENODATA:
            return xt
        if xt == ENODATA or xt < self.volmark:
            if opts['create']:
                xt = _xtime_now()
                rsc.server.aggregated.set_xtime(path, self.uuid, xt)
            else:
                zero_zero = (0, 0)
                if xt != zero_zero:
                    xt = opts['default_xtime']
        return xt

    def keepalive_payload_hook(self, timo, gap):
        # first grab a reference as self.volinfo
        # can be changed in main thread
        vi = self.volinfo
        if vi:
            # then have a private copy which we can mod
            vi = vi.copy()
            vi['timeout'] = int(time.time()) + timo
        else:
            # send keep-alive more frequently to
            # avoid a delay in announcing our volume info
            # to slave if it becomes established in the
            # meantime
            gap = min(10, gap)
        return (vi, gap)

    def volinfo_hook(self):
        return self.get_sys_volinfo()

    def xtime_reversion_hook(self, path, xtl, xtr):
        if xtr > xtl:
            raise GsyncdError("timestamp corruption for " + path)

    def need_sync(self, e, xte, xtrd):
        return xte > xtrd

    def set_slave_xtime(self, path, mark):
        self.slave.server.set_stime(path, self.uuid, mark)
        # self.slave.server.set_xtime_remote(path, self.uuid, mark)


class PartialMixin(NormalMixin):

    """a variant tuned towards operation with a master
       that has partial info of the slave (brick typically)"""

    def xtime_reversion_hook(self, path, xtl, xtr):
        pass


class RecoverMixin(NormalMixin):

    """a variant that differs from normal in terms
       of ignoring non-indexed files"""

    @staticmethod
    def make_xtime_opts(is_master, opts):
        if 'create' not in opts:
            opts['create'] = False
        if 'default_xtime' not in opts:
            opts['default_xtime'] = URXTIME

    def keepalive_payload_hook(self, timo, gap):
        return (None, gap)

    def volinfo_hook(self):
        return _volinfo_hook_relax_foreign(self)

# Further mixins for certain tunable behaviors


class SendmarkNormalMixin(object):

    def sendmark_regular(self, *a, **kw):
        return self.sendmark(*a, **kw)


class SendmarkRsyncMixin(object):

    def sendmark_regular(self, *a, **kw):
        pass


class PurgeNormalMixin(object):

    def purge_missing(self, path, names):
        self.slave.server.purge(path, names)


class PurgeNoopMixin(object):

    def purge_missing(self, path, names):
        pass


class TarSSHEngine(object):

    """Sync engine that uses tar(1) piped over ssh(1)
       for data transfers. Good for lots of small files.
    """

    def a_syncdata(self, files):
        logging.debug(lf("Files", files=files))

        for f in files:
            pb = self.syncer.add(f)

            def regjob(se, xte, pb):
                rv = pb.wait()
                if rv[0]:
                    logging.debug(lf('synced', file=se))
                    return True
                else:
                    # stat check for file presence
                    st = lstat(se)
                    if isinstance(st, int):
                        # file got unlinked in the interim
                        self.unlinked_gfids.add(se)
                        return True

            self.add_job(self.FLAT_DIR_HIERARCHY, 'reg', regjob, f, None, pb)

    def syncdata_wait(self):
        if self.wait(self.FLAT_DIR_HIERARCHY, None):
            return True

    def syncdata(self, files):
        self.a_syncdata(files)
        self.syncdata_wait()


class RsyncEngine(object):

    """Sync engine that uses rsync(1) for data transfers"""

    def a_syncdata(self, files):
        logging.debug(lf("files", files=files))

        for f in files:
            logging.debug(lf('candidate for syncing', file=f))
            pb = self.syncer.add(f)

            def regjob(se, xte, pb):
                rv = pb.wait()
                if rv[0]:
                    logging.debug(lf('synced', file=se))
                    return True
                else:
                    # stat to check if the file exist
                    st = lstat(se)
                    if isinstance(st, int):
                        # file got unlinked in the interim
                        self.unlinked_gfids.add(se)
                        return True

            self.add_job(self.FLAT_DIR_HIERARCHY, 'reg', regjob, f, None, pb)

    def syncdata_wait(self):
        if self.wait(self.FLAT_DIR_HIERARCHY, None):
            return True

    def syncdata(self, files):
        self.a_syncdata(files)
        self.syncdata_wait()


class GMasterCommon(object):

    """abstract class impementling master role"""

    KFGN = 0
    KNAT = 1

    def get_sys_volinfo(self):
        """query volume marks on fs root

        err out on multiple foreign masters
        """
        fgn_vis, nat_vi = (
            self.master.server.aggregated.foreign_volume_infos(),
            self.master.server.aggregated.native_volume_info())
        fgn_vi = None
        if fgn_vis:
            if len(fgn_vis) > 1:
                raise GsyncdError("cannot work with multiple foreign masters")
            fgn_vi = fgn_vis[0]
        return fgn_vi, nat_vi

    @property
    def uuid(self):
        if self.volinfo:
            return self.volinfo['uuid']

    @property
    def volmark(self):
        if self.volinfo:
            return self.volinfo['volume_mark']

    def get_entry_stime(self):
        data = self.slave.server.entry_stime(".", self.uuid)
        if isinstance(data, int):
            data = None
        return data

    def get_data_stime(self):
        data = self.slave.server.stime(".", self.uuid)
        if isinstance(data, int):
            data = None
        return data

    def xtime(self, path, *a, **opts):
        """get amended xtime

        as of amending, we can create missing xtime, or
        determine a valid value if what we get is expired
        (as of the volume mark expiry); way of amendig
        depends on @opts and on subject of query (master
        or slave).
        """
        if a:
            rsc = a[0]
        else:
            rsc = self.master
        self.make_xtime_opts(rsc == self.master, opts)
        return self.xtime_low(rsc, path, **opts)

    def __init__(self, master, slave):
        self.master = master
        self.slave = slave
        self.jobtab = {}
        if gconf.get("sync-method") == "tarssh":
            self.syncer = Syncer(slave, self.slave.tarssh, [2])
        else:
            # partial transfer (cf. rsync(1)), that's normal
            self.syncer = Syncer(slave, self.slave.rsync, [23, 24])
        # crawls vs. turns:
        # - self.crawls is simply the number of crawl() invocations on root
        # - one turn is a maximal consecutive sequence of crawls so that each
        #   crawl in it detects a change to be synced
        # - self.turns is the number of turns since start
        # - self.total_turns is a limit so that if self.turns reaches it, then
        #   we exit (for diagnostic purposes)
        # so, eg., if the master fs changes unceasingly, self.turns will remain
        # 0.
        self.crawls = 0
        self.turns = 0
        self.total_turns = rconf.turns
        self.crawl_start = datetime.now()
        self.lastreport = {'crawls': 0, 'turns': 0, 'time': 0}
        self.start = None
        self.change_seen = None
        # the actual volinfo we make use of
        self.volinfo = None
        self.terminate = False
        self.sleep_interval = 1
        self.unlinked_gfids = set()

    def init_keep_alive(cls):
        """start the keep-alive thread """
        timo = gconf.get("slave-timeout", 0)
        if timo > 0:
            def keep_alive():
                while True:
                    vi, gap = cls.keepalive_payload_hook(timo, timo * 0.5)
                    cls.slave.server.keep_alive(vi)
                    time.sleep(gap)
            t = Thread(target=keep_alive)
            t.start()

    def mgmt_lock(self):

        """Take management volume lock """
        if rconf.mgmt_lock_fd:
            try:
                fcntl.lockf(rconf.mgmt_lock_fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
                return True
            except:
                ex = sys.exc_info()[1]
                if isinstance(ex, IOError) and ex.errno in (EACCES, EAGAIN):
                    return False
                raise

        fd = None
        bname = str(self.uuid) + "_" + rconf.args.slave_id + "_subvol_" \
            + str(rconf.args.subvol_num) + ".lock"
        mgmt_lock_dir = os.path.join(gconf.get("meta-volume-mnt"), "geo-rep")
        path = os.path.join(mgmt_lock_dir, bname)
        logging.debug(lf("lock file path", path=path))
        try:
            fd = os.open(path, os.O_CREAT | os.O_RDWR)
        except OSError:
            ex = sys.exc_info()[1]
            if ex.errno == ENOENT:
                logging.info("Creating geo-rep directory in meta volume...")
                try:
                    os.makedirs(mgmt_lock_dir)
                except OSError:
                    ex = sys.exc_info()[1]
                    if ex.errno == EEXIST:
                        pass
                    else:
                        raise
                fd = os.open(path, os.O_CREAT | os.O_RDWR)
            else:
                raise
        try:
            fcntl.lockf(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
            # Save latest FD for future use
            rconf.mgmt_lock_fd = fd
        except:
            ex = sys.exc_info()[1]
            if isinstance(ex, IOError) and ex.errno in (EACCES, EAGAIN):
                # cannot grab, it's taken
                rconf.mgmt_lock_fd = fd
                return False
            raise

        return True

    def should_crawl(self):
        if not gconf.get("use-meta-volume"):
            return rconf.args.local_node_id in self.master.server.node_uuid()

        if not os.path.ismount(gconf.get("meta-volume-mnt")):
            logging.error("Meta-volume is not mounted. Worker Exiting...")
            sys.exit(1)
        return self.mgmt_lock()

    def register(self):
        self.register()

    def crawlwrap(self, oneshot=False, register_time=None):
        if oneshot:
            # it's important to do this during the oneshot crawl as
            # for a passive gsyncd (ie. in a replicate scenario)
            # the keepalive thread would keep the connection alive.
            self.init_keep_alive()

        # If crawlwrap is called when partial history available,
        # then it sets register_time which is the time when geo-rep
        # worker registered to changelog consumption. Since nsec is
        # not considered in register time, there are chances of skipping
        # changes detection in xsync crawl. This limit will be reset when
        # crawlwrap is called again.
        self.live_changelog_start_time = None
        if register_time:
            self.live_changelog_start_time = (register_time, 0)

        # no need to maintain volinfo state machine.
        # in a cascading setup, each geo-replication session is
        # independent (ie. 'volume-mark' and 'xtime' are not
        # propagated). This is because the slave's xtime is now
        # stored on the master itself. 'volume-mark' just identifies
        # that we are in a cascading setup and need to enable
        # 'geo-replication.ignore-pid-check' option.
        volinfo_sys = self.volinfo_hook()
        self.volinfo = volinfo_sys[self.KNAT]
        inter_master = volinfo_sys[self.KFGN]
        logging.debug("%s master with volume id %s ..." %
                      (inter_master and "intermediate" or "primary",
                       self.uuid))
        rconf.volume_id = self.uuid
        if self.volinfo:
            if self.volinfo['retval']:
                logging.warn(lf("master cluster's info may not be valid",
                                error=self.volinfo['retval']))
        else:
            raise GsyncdError("master volinfo unavailable")
        self.lastreport['time'] = time.time()

        t0 = time.time()
        crawl = self.should_crawl()
        while not self.terminate:
            if self.start:
                logging.debug("... crawl #%d done, took %.6f seconds" %
                              (self.crawls, time.time() - self.start))
            self.start = time.time()
            should_display_info = self.start - self.lastreport['time'] >= 60
            if should_display_info:
                logging.debug("%d crawls, %d turns",
                              self.crawls - self.lastreport['crawls'],
                              self.turns - self.lastreport['turns'])
                self.lastreport.update(crawls=self.crawls,
                                       turns=self.turns,
                                       time=self.start)
            t1 = time.time()
            if int(t1 - t0) >= gconf.get("replica-failover-interval"):
                crawl = self.should_crawl()
                t0 = t1
            self.update_worker_remote_node()
            if not crawl:
                self.status.set_passive()
                # bring up _this_ brick to the cluster stime
                # which is min of cluster (but max of the replicas)
                brick_stime = self.xtime('.', self.slave)
                cluster_stime = self.master.server.aggregated.stime_mnt(
                    '.', '.'.join([str(self.uuid), rconf.args.slave_id]))
                logging.debug(lf("Crawl info",
                                 cluster_stime=cluster_stime,
                                 brick_stime=brick_stime))

                if not isinstance(cluster_stime, int):
                    if brick_stime < cluster_stime:
                        self.slave.server.set_stime(
                            self.FLAT_DIR_HIERARCHY, self.uuid, cluster_stime)
                        self.upd_stime(cluster_stime)
                        # Purge all changelogs available in processing dir
                        # less than cluster_stime
                        proc_dir = os.path.join(self.tempdir,
                                                ".processing")

                        if os.path.exists(proc_dir):
                            to_purge = [f for f in os.listdir(proc_dir)
                                        if (f.startswith("CHANGELOG.") and
                                            int(f.split('.')[-1]) <
                                            cluster_stime[0])]
                            for f in to_purge:
                                os.remove(os.path.join(proc_dir, f))

                time.sleep(5)
                continue

            self.status.set_active()
            self.crawl()

            if oneshot:
                return
            time.sleep(self.sleep_interval)

    @staticmethod
    def humantime(*tpair):
        """format xtime-like (sec, nsec) pair to human readable format"""
        ts = datetime.fromtimestamp(float('.'.join(str(n) for n in tpair))).\
            strftime("%Y-%m-%d %H:%M:%S")
        if len(tpair) > 1:
            ts += '.' + str(tpair[1])
        return ts

    def _crawl_time_format(self, crawl_time):
        # Ex: 5 years, 4 days, 20:23:10
        years, days = divmod(crawl_time.days, 365.25)
        years = int(years)
        days = int(days)

        date = ""
        m, s = divmod(crawl_time.seconds, 60)
        h, m = divmod(m, 60)

        if years != 0:
            date += "%s %s " % (years, "year" if years == 1 else "years")
        if days != 0:
            date += "%s %s " % (days, "day" if days == 1 else "days")

        date += "%s:%s:%s" % (string.zfill(h, 2),
                              string.zfill(m, 2), string.zfill(s, 2))
        return date

    def add_job(self, path, label, job, *a, **kw):
        """insert @job function to job table at @path with @label"""
        if self.jobtab.get(path) is None:
            self.jobtab[path] = []
        self.jobtab[path].append((label, a, lambda: job(*a, **kw)))

    def add_failjob(self, path, label):
        """invoke .add_job with a job that does nothing just fails"""
        logging.debug('salvaged: ' + label)
        self.add_job(path, label, lambda: False)

    def wait(self, path, *args):
        """perform jobs registered for @path

        Reset jobtab entry for @path,
        determine success as the conjunction of
        success of all the jobs. In case of
        success, call .sendmark on @path
        """
        jobs = self.jobtab.pop(path, [])
        succeed = True
        for j in jobs:
            ret = j[-1]()
            if not ret:
                succeed = False
        if succeed and not args[0] is None:
            self.sendmark(path, *args)
        return succeed

    def sendmark(self, path, mark, adct=None):
        """update slave side xtime for @path to master side xtime

        also can send a setattr payload (see Server.setattr).
        """
        if adct:
            self.slave.server.setattr(path, adct)
        self.set_slave_xtime(path, mark)


class XCrawlMetadata(object):
    def __init__(self, st_uid, st_gid, st_mode, st_atime, st_mtime):
        self.st_uid = int(st_uid)
        self.st_gid = int(st_gid)
        self.st_mode = int(st_mode)
        self.st_atime = float(st_atime)
        self.st_mtime = float(st_mtime)


class GMasterChangelogMixin(GMasterCommon):

    """ changelog based change detection and syncing """

    # index for change type and entry
    IDX_START = 0
    IDX_END = 2
    UNLINK_ENTRY = 2

    POS_GFID = 0
    POS_TYPE = 1
    POS_ENTRY1 = -1

    TYPE_META = "M "
    TYPE_GFID = "D "
    TYPE_ENTRY = "E "

    MAX_EF_RETRIES = 10
    MAX_OE_RETRIES = 10

    # flat directory hierarchy for gfid based access
    FLAT_DIR_HIERARCHY = '.'

    CHANGELOG_CONN_RETRIES = 5

    def init_fop_batch_stats(self):
        self.batch_stats = {
            "CREATE": 0,
            "MKNOD": 0,
            "UNLINK": 0,
            "MKDIR": 0,
            "RMDIR": 0,
            "LINK": 0,
            "SYMLINK": 0,
            "RENAME": 0,
            "SETATTR": 0,
            "SETXATTR": 0,
            "XATTROP": 0,
            "DATA": 0,
            "ENTRY_SYNC_TIME": 0,
            "META_SYNC_TIME": 0,
            "DATA_START_TIME": 0
        }

    def update_fop_batch_stats(self, ty):
        if ty in ['FSETXATTR']:
            ty = 'SETXATTR'
        self.batch_stats[ty] = self.batch_stats.get(ty, 0) + 1

    def archive_and_purge_changelogs(self, changelogs):
        # Creates tar file instead of tar.gz, since changelogs will
        # be appended to existing tar. archive name is
        # archive_<YEAR><MONTH>.tar
        archive_name = "archive_%s.tar" % datetime.today().strftime(
            gconf.get("changelog-archive-format"))

        try:
            tar = tarfile.open(os.path.join(self.processed_changelogs_dir,
                                            archive_name),
                               "a")
        except tarfile.ReadError:
            tar = tarfile.open(os.path.join(self.processed_changelogs_dir,
                                            archive_name),
                               "w")

        for f in changelogs:
            try:
                f = os.path.basename(f)
                tar.add(os.path.join(self.processed_changelogs_dir, f),
                        arcname=os.path.basename(f))
            except:
                exc = sys.exc_info()[1]
                if ((isinstance(exc, OSError) or
                     isinstance(exc, IOError)) and exc.errno == ENOENT):
                    continue
                else:
                    tar.close()
                    raise
        tar.close()

        for f in changelogs:
            try:
                f = os.path.basename(f)
                os.remove(os.path.join(self.processed_changelogs_dir, f))
            except OSError as e:
                if e.errno == errno.ENOENT:
                    continue
                else:
                    raise

    def setup_working_dir(self):
        workdir = os.path.join(gconf.get("working-dir"),
                               escape(rconf.args.local_path))
        logging.debug('changelog working dir %s' % workdir)
        return workdir

    def log_failures(self, failures, entry_key, gfid_prefix, log_prefix):
        num_failures = 0
        for failure in failures:
            st = lstat(os.path.join(gfid_prefix, failure[0][entry_key]))
            if not isinstance(st, int):
                num_failures += 1
                logging.error(lf('%s FAILED' % log_prefix,
                                 data=failure))
                if failure[0]['op'] == 'MKDIR':
                    raise GsyncdError("The above directory failed to sync."
                                      " Please fix it to proceed further.")

        self.status.inc_value("failures", num_failures)

    def fix_possible_entry_failures(self, failures, retry_count, entries):
        pfx = gauxpfx()
        fix_entry_ops = []
        failures1 = []
        remove_gfids = set()
        for failure in failures:
            if failure[2]['name_mismatch']:
                pbname = failure[2]['slave_entry']
            elif failure[2]['dst']:
                pbname = failure[0]['entry1']
            else:
                pbname = failure[0]['entry']

            op = failure[0]['op']
            # name exists but gfid is different
            if failure[2]['gfid_mismatch'] or failure[2]['name_mismatch']:
                slave_gfid = failure[2]['slave_gfid']
                st = lstat(os.path.join(pfx, slave_gfid))
                # Takes care of scenarios with no hardlinks
                if isinstance(st, int) and st == ENOENT:
                    logging.debug(lf('Entry not present on master. Fixing gfid '
                                    'mismatch in slave. Deleting the entry',
                                    retry_count=retry_count,
                                    entry=repr(failure)))
                    # Add deletion to fix_entry_ops list
                    if failure[2]['slave_isdir']:
                        fix_entry_ops.append(
                            edct('RMDIR',
                                 gfid=failure[2]['slave_gfid'],
                                 entry=pbname))
                    else:
                        fix_entry_ops.append(
                            edct('UNLINK',
                                 gfid=failure[2]['slave_gfid'],
                                 entry=pbname))
                    remove_gfids.add(slave_gfid)
                    if op in ['RENAME']:
                        # If renamed gfid doesn't exists on master, remove
                        # rename entry and unlink src on slave
                        st = lstat(os.path.join(pfx, failure[0]['gfid']))
                        if isinstance(st, int) and st == ENOENT:
                            logging.debug("Unlink source %s" % repr(failure))
                            remove_gfids.add(failure[0]['gfid'])
                            fix_entry_ops.append(
                                edct('UNLINK',
                                     gfid=failure[0]['gfid'],
                                     entry=failure[0]['entry']))
                # Takes care of scenarios of hardlinks/renames on master
                elif not isinstance(st, int):
                    if matching_disk_gfid(slave_gfid, pbname):
                        # Safe to ignore the failure as master contains same
                        # file with same gfid. Remove entry from entries list
                        logging.debug(lf('Fixing gfid mismatch in slave. '
                                        ' Safe to ignore, take out entry',
                                        retry_count=retry_count,
                                        entry=repr(failure)))
                        remove_gfids.add(failure[0]['gfid'])
                        if op == 'RENAME':
                            fix_entry_ops.append(
                                edct('UNLINK',
                                     gfid=failure[0]['gfid'],
                                     entry=failure[0]['entry']))
                    # The file exists on master but with different name.
                    # Probably renamed and got missed during xsync crawl.
                    elif failure[2]['slave_isdir']:
                        realpath = os.readlink(os.path.join(
                                               rconf.args.local_path,
                                               ".glusterfs",
                                               slave_gfid[0:2],
                                               slave_gfid[2:4],
                                               slave_gfid))
                        dst_entry = os.path.join(pfx, realpath.split('/')[-2],
                                                 realpath.split('/')[-1])
                        src_entry = pbname
                        logging.debug(lf('Fixing dir name/gfid mismatch in '
                                        'slave', retry_count=retry_count,
                                        entry=repr(failure)))
                        if src_entry == dst_entry:
                            # Safe to ignore the failure as master contains
                            # same directory as in slave with same gfid.
                            # Remove the failure entry from entries list
                            logging.debug(lf('Fixing dir name/gfid mismatch'
                                            ' in slave. Safe to ignore, '
                                            'take out entry',
                                            retry_count=retry_count,
                                            entry=repr(failure)))
                            try:
                                entries.remove(failure[0])
                            except ValueError:
                                pass
                        else:
                            rename_dict = edct('RENAME', gfid=slave_gfid,
                                               entry=src_entry,
                                               entry1=dst_entry, stat=st,
                                               link=None)
                            logging.debug(lf('Fixing dir name/gfid mismatch'
                                            ' in slave. Renaming',
                                            retry_count=retry_count,
                                            entry=repr(rename_dict)))
                            fix_entry_ops.append(rename_dict)
                    else:
                        # A hardlink file exists with different name or
                        # renamed file exists and we are sure from
                        # matching_disk_gfid check that the entry doesn't
                        # exist with same gfid so we can safely delete on slave
                        logging.debug(lf('Fixing file gfid mismatch in slave. '
                                        'Hardlink/Rename Case. Deleting entry',
                                        retry_count=retry_count,
                                        entry=repr(failure)))
                        fix_entry_ops.append(
                            edct('UNLINK',
                                 gfid=failure[2]['slave_gfid'],
                                 entry=pbname))
            elif failure[1] == ENOENT:
                if op in ['RENAME']:
                    pbname = failure[0]['entry1']
                else:
                    pbname = failure[0]['entry']

                pargfid = pbname.split('/')[1]
                st = lstat(os.path.join(pfx, pargfid))
                # Safe to ignore the failure as master doesn't contain
                # parent directory.
                if isinstance(st, int):
                    logging.debug(lf('Fixing ENOENT error in slave. Parent '
                                    'does not exist on master. Safe to '
                                    'ignore, take out entry',
                                    retry_count=retry_count,
                                    entry=repr(failure)))
                    try:
                        entries.remove(failure[0])
                    except ValueError:
                        pass
                else:
                    logging.debug(lf('Fixing ENOENT error in slave. Create '
                                    'parent directory on slave.',
                                    retry_count=retry_count,
                                    entry=repr(failure)))
                    realpath = os.readlink(os.path.join(rconf.args.local_path,
                                                        ".glusterfs",
                                                        pargfid[0:2],
                                                        pargfid[2:4],
                                                        pargfid))
                    dir_entry = os.path.join(pfx, realpath.split('/')[-2],
                                             realpath.split('/')[-1])
                    fix_entry_ops.append(
                        edct('MKDIR', gfid=pargfid, entry=dir_entry,
                             mode=st.st_mode, uid=st.st_uid, gid=st.st_gid))

        logging.debug("remove_gfids: %s" % repr(remove_gfids))
        if remove_gfids:
            for e in entries:
                if e['op'] in ['MKDIR', 'MKNOD', 'CREATE', 'RENAME'] \
                   and e['gfid'] in remove_gfids:
                    logging.debug("Removed entry op from retrial list: entry: %s" % repr(e))
                    e['skip_entry'] = True

        if fix_entry_ops:
            # Process deletions of entries whose gfids are mismatched
            failures1 = self.slave.server.entry_ops(fix_entry_ops)

        return (failures1, fix_entry_ops)

    def handle_entry_failures(self, failures, entries):
        retries = 0
        pending_failures = False
        failures1 = []
        failures2 = []
        entry_ops1 = []
        entry_ops2 = []

        if failures:
            pending_failures = True
            failures1 = failures
            entry_ops1 = entries

            while pending_failures and retries < self.MAX_EF_RETRIES:
                retries += 1
                (failures2, entry_ops2) = self.fix_possible_entry_failures(
                    failures1, retries, entry_ops1)
                if not failures2:
                    pending_failures = False
                    logging.info(lf('Successfully fixed entry ops with gfid '
                                 'mismatch', retry_count=retries))
                else:
                    pending_failures = True
                    failures1 = failures2
                    entry_ops1 = entry_ops2

            if pending_failures:
                for failure in failures1:
                    logging.error("Failed to fix entry ops %s", repr(failure))

    def process_change(self, change, done, retry):
        pfx = gauxpfx()
        clist = []
        entries = []
        meta_gfid = set()
        datas = set()

        change_ts = change.split(".")[-1]

        # Ignore entry ops which are already processed in Changelog modes
        ignore_entry_ops = False
        entry_stime = None
        data_stime = None
        if self.name in ["live_changelog", "history_changelog"]:
            entry_stime = self.get_entry_stime()
            data_stime = self.get_data_stime()

        if entry_stime is not None and data_stime is not None:
            # if entry_stime is not None but data_stime > entry_stime
            # This situation is caused by the stime update of Passive worker
            # Consider data_stime in this case.
            if data_stime[0] > entry_stime[0]:
                entry_stime = data_stime

            # Compare the entry_stime with changelog file suffix
            # if changelog time is less than entry_stime then ignore
            if int(change_ts) <= entry_stime[0]:
                ignore_entry_ops = True

        try:
            f = open(change, "r")
            clist = f.readlines()
            f.close()
        except IOError:
            raise

        for e in clist:
            e = e.strip()
            et = e[self.IDX_START:self.IDX_END]   # entry type
            ec = e[self.IDX_END:].split(' ')      # rest of the bits

            # skip ENTRY operation if hot tier brick
            if self.name == 'live_changelog' or \
               self.name == 'history_changelog':
                if rconf.args.is_hottier and et == self.TYPE_ENTRY:
                    logging.debug(lf('skip ENTRY op if hot tier brick',
                                     op=ec[self.POS_TYPE]))
                    continue

            # Data and Meta operations are decided while parsing
            # UNLINK/RMDIR/MKNOD except that case ignore all the other
            # entry ops if ignore_entry_ops is True.
            # UNLINK/RMDIR/MKNOD entry_ops are ignored in the end
            if ignore_entry_ops and et == self.TYPE_ENTRY and \
               ec[self.POS_TYPE] not in ["UNLINK", "RMDIR", "MKNOD"]:
                continue

            if et == self.TYPE_ENTRY:
                # extract information according to the type of
                # the entry operation. create(), mkdir() and mknod()
                # have mode, uid, gid information in the changelog
                # itself, so no need to stat()...
                ty = ec[self.POS_TYPE]

                self.update_fop_batch_stats(ec[self.POS_TYPE])

                # PARGFID/BNAME
                en = unescape_space_newline(
                    os.path.join(pfx, ec[self.POS_ENTRY1]))
                # GFID of the entry
                gfid = ec[self.POS_GFID]

                if ty in ['UNLINK', 'RMDIR']:
                    # The index of PARGFID/BNAME for UNLINK, RMDIR
                    # is no more the last index. It varies based on
                    # changelog.capture-del-path is enabled or not.
                    en = unescape_space_newline(
                        os.path.join(pfx, ec[self.UNLINK_ENTRY]))

                    # Remove from DATA list, so that rsync will
                    # not fail
                    pt = os.path.join(pfx, ec[0])
                    st = lstat(pt)
                    if pt in datas and isinstance(st, int):
                        # file got unlinked, May be historical Changelog
                        datas.remove(pt)

                    if ty in ['RMDIR'] and not isinstance(st, int):
                        logging.info(lf('Ignoring rmdir. Directory present in '
                                        'master', gfid=gfid, pgfid_bname=en))
                        continue

                    if not gconf.get("ignore-deletes"):
                        if not ignore_entry_ops:
                            entries.append(edct(ty, gfid=gfid, entry=en))
                elif ty in ['CREATE', 'MKDIR', 'MKNOD']:
                    # Special case: record mknod as link
                    if ty in ['MKNOD']:
                        mode = int(ec[2])
                        if mode & 0o1000:
                                # Avoid stat'ing the file as it
                                # may be deleted in the interim
                                st = FreeObject(st_mode=int(ec[2]),
                                                st_uid=int(ec[3]),
                                                st_gid=int(ec[4]),
                                                st_atime=0,
                                                st_mtime=0)

                                # So, it may be deleted, but still we are
                                # append LINK? Because, the file will be
                                # CREATED if source not exists.
                                entries.append(edct('LINK', stat=st, entry=en,
                                               gfid=gfid))

                                # Here, we have the assumption that only
                                # tier-gfid.linkto causes this mknod. Add data
                                datas.add(os.path.join(pfx, ec[0]))
                                continue

                    # stat info. present in the changelog itself
                    entries.append(edct(ty, gfid=gfid, entry=en,
                                   mode=int(ec[2]),
                                   uid=int(ec[3]), gid=int(ec[4])))
                elif ty == "RENAME":
                    go = os.path.join(pfx, gfid)
                    st = lstat(go)
                    if isinstance(st, int):
                        st = {}

                    rl = None
                    if st and stat.S_ISLNK(st.st_mode):
                        rl = errno_wrap(os.readlink, [en], [ENOENT],
                                        [ESTALE, EINTR])
                        if isinstance(rl, int):
                            rl = None

                    e1 = unescape_space_newline(
                        os.path.join(pfx, ec[self.POS_ENTRY1 - 1]))
                    entries.append(edct(ty, gfid=gfid, entry=e1, entry1=en,
                                        stat=st, link=rl))
                    # If src doesn't exist while doing rename, destination
                    # is created. If data is not followed by rename, this
                    # remains zero byte file on slave. Hence add data entry
                    # for renames
                    datas.add(os.path.join(pfx, gfid))
                else:
                    # stat() to get mode and other information
                    if not matching_disk_gfid(gfid, en):
                        logging.debug(lf('Ignoring entry, purged in the '
                                      'interim', file=en, gfid=gfid))
                        continue

                    go = os.path.join(pfx, gfid)
                    st = lstat(go)
                    if isinstance(st, int):
                        logging.debug(lf('Ignoring entry, purged in the '
                                      'interim', file=en, gfid=gfid))
                        continue

                    if ty == 'LINK':
                        rl = None
                        if st and stat.S_ISLNK(st.st_mode):
                            rl = errno_wrap(os.readlink, [en], [ENOENT],
                                            [ESTALE, EINTR])
                            if isinstance(rl, int):
                                rl = None
                        entries.append(edct(ty, stat=st, entry=en, gfid=gfid,
                                       link=rl))
                        # If src doesn't exist while doing link, destination
                        # is created based on file type. If data is not
                        # followed by link, this remains zero byte file on
                        # slave. Hence add data entry for links
                        if rl is None:
                            datas.add(os.path.join(pfx, gfid))
                    elif ty == 'SYMLINK':
                        rl = errno_wrap(os.readlink, [en], [ENOENT],
                                        [ESTALE, EINTR])
                        if isinstance(rl, int):
                            continue

                        entries.append(
                            edct(ty, stat=st, entry=en, gfid=gfid, link=rl))
                    else:
                        logging.warn(lf('ignoring op',
                                        gfid=gfid,
                                        type=ty))
            elif et == self.TYPE_GFID:
                # If self.unlinked_gfids is available, then that means it is
                # retrying the changelog second time. Do not add the GFID's
                # to rsync job if failed previously but unlinked in master
                if self.unlinked_gfids and \
                   os.path.join(pfx, ec[0]) in self.unlinked_gfids:
                    logging.debug("ignoring data, since file purged interim")
                else:
                    datas.add(os.path.join(pfx, ec[0]))
            elif et == self.TYPE_META:
                self.update_fop_batch_stats(ec[self.POS_TYPE])
                if ec[1] == 'SETATTR':  # only setattr's for now...
                    if len(ec) == 5:
                        # In xsync crawl, we already have stat data
                        # avoid doing stat again
                        meta_gfid.add((os.path.join(pfx, ec[0]),
                                       XCrawlMetadata(st_uid=ec[2],
                                                      st_gid=ec[3],
                                                      st_mode=ec[4],
                                                      st_atime=ec[5],
                                                      st_mtime=ec[6])))
                    else:
                        meta_gfid.add((os.path.join(pfx, ec[0]), ))
                elif ec[1] in ['SETXATTR', 'XATTROP', 'FXATTROP']:
                    # To sync xattr/acls use rsync/tar, --xattrs and --acls
                    # switch to rsync and tar
                    if not gconf.get("sync-method") == "tarssh" and \
                       (gconf.get("sync-xattrs") or gconf.get("sync-acls")):
                        datas.add(os.path.join(pfx, ec[0]))
            else:
                logging.warn(lf('got invalid fop type',
                                type=et))
        logging.debug('entries: %s' % repr(entries))

        # Increment counters for Status
        self.files_in_batch += len(datas)
        self.status.inc_value("data", len(datas))

        self.batch_stats["DATA"] += self.files_in_batch - \
            self.batch_stats["SETXATTR"] - \
            self.batch_stats["XATTROP"]

        entry_start_time = time.time()
        # sync namespace
        if entries and not ignore_entry_ops:
            # Increment counters for Status
            self.status.inc_value("entry", len(entries))

            failures = self.slave.server.entry_ops(entries)

            if gconf.get("gfid-conflict-resolution"):
                count = 0
                if failures:
                    logging.info(lf('Entry ops failed with gfid mismatch',
                                count=len(failures)))
                while failures and count < self.MAX_OE_RETRIES:
                    count += 1
                    self.handle_entry_failures(failures, entries)
                    logging.info(lf('Retry original entries', count=count))
                    failures = self.slave.server.entry_ops(entries)
                    if not failures:
                        logging.info("Successfully fixed all entry ops with "
                                     "gfid mismatch")
                        break

            self.log_failures(failures, 'gfid', gauxpfx(), 'ENTRY')
            self.status.dec_value("entry", len(entries))

            # Update Entry stime in Brick Root only in case of Changelog mode
            if self.name in ["live_changelog", "history_changelog"]:
                entry_stime_to_update = (int(change_ts) - 1, 0)
                self.upd_entry_stime(entry_stime_to_update)
                self.status.set_field("last_synced_entry",
                                      entry_stime_to_update[0])

        self.batch_stats["ENTRY_SYNC_TIME"] += time.time() - entry_start_time

        if ignore_entry_ops:
            # Book keeping, to show in logs the range of Changelogs skipped
            self.num_skipped_entry_changelogs += 1
            if self.skipped_entry_changelogs_first is None:
                self.skipped_entry_changelogs_first = change_ts

            self.skipped_entry_changelogs_last = change_ts

        meta_start_time = time.time()
        # sync metadata
        if meta_gfid:
            meta_entries = []
            for go in meta_gfid:
                if len(go) > 1:
                    st = go[1]
                else:
                    st = lstat(go[0])
                if isinstance(st, int):
                    logging.debug(lf('file got purged in the interim',
                                     file=go[0]))
                    continue
                meta_entries.append(edct('META', go=go[0], stat=st))
            if meta_entries:
                self.status.inc_value("meta", len(meta_entries))
                failures = self.slave.server.meta_ops(meta_entries)
                self.log_failures(failures, 'go', '', 'META')
                self.status.dec_value("meta", len(meta_entries))

        self.batch_stats["META_SYNC_TIME"] += time.time() - meta_start_time

        if self.batch_stats["DATA_START_TIME"] == 0:
            self.batch_stats["DATA_START_TIME"] = time.time()

        # sync data
        if datas:
            self.a_syncdata(datas)
            self.datas_in_batch.update(datas)

    def process(self, changes, done=1):
        tries = 0
        retry = False
        self.unlinked_gfids = set()
        self.files_in_batch = 0
        self.datas_in_batch = set()
        # Error log disabled till the last round
        self.syncer.disable_errorlog()
        self.skipped_entry_changelogs_first = None
        self.skipped_entry_changelogs_last = None
        self.num_skipped_entry_changelogs = 0
        self.batch_start_time = time.time()
        self.init_fop_batch_stats()

        while True:
            # first, fire all changelog transfers in parallel. entry and
            # metadata are performed synchronously, therefore in serial.
            # However at the end of each changelog, data is synchronized
            # with syncdata_async() - which means it is serial w.r.t
            # entries/metadata of that changelog but happens in parallel
            # with data of other changelogs.

            if retry:
                if tries == (gconf.get("max-rsync-retries") - 1):
                    # Enable Error logging if it is last retry
                    self.syncer.enable_errorlog()

                # Remove Unlinked GFIDs from Queue
                for unlinked_gfid in self.unlinked_gfids:
                    if unlinked_gfid in self.datas_in_batch:
                        self.datas_in_batch.remove(unlinked_gfid)

                # Retry only Sync. Do not retry entry ops
                if self.datas_in_batch:
                    self.a_syncdata(self.datas_in_batch)
            else:
                for change in changes:
                    logging.debug(lf('processing change',
                                     changelog=change))
                    self.process_change(change, done, retry)
                    if not retry:
                        # number of changelogs processed in the batch
                        self.turns += 1

            # Now we wait for all the data transfers fired off in the above
            # step to complete. Note that this is not ideal either. Ideally
            # we want to trigger the entry/meta-data transfer of the next
            # batch while waiting for the data transfer of the current batch
            # to finish.

            # Note that the reason to wait for the data transfer (vs doing it
            # completely in the background and call the changelog_done()
            # asynchronously) is because this waiting acts as a "backpressure"
            # and prevents a spiraling increase of wait stubs from consuming
            # unbounded memory and resources.

            # update the slave's time with the timestamp of the _last_
            # changelog file time suffix. Since, the changelog prefix time
            # is the time when the changelog was rolled over, introduce a
            # tolerance of 1 second to counter the small delta b/w the
            # marker update and gettimeofday().
            # NOTE: this is only for changelog mode, not xsync.

            # @change is the last changelog (therefore max time for this batch)
            if self.syncdata_wait():
                self.unlinked_gfids = set()
                if done:
                    xtl = (int(change.split('.')[-1]) - 1, 0)
                    self.upd_stime(xtl)
                    list(map(self.changelog_done_func, changes))
                    self.archive_and_purge_changelogs(changes)

                # Reset Data counter after sync
                self.status.dec_value("data", self.files_in_batch)
                self.files_in_batch = 0
                self.datas_in_batch = set()
                break

            # We do not know which changelog transfer failed, retry everything.
            retry = True
            tries += 1
            if tries == gconf.get("max-rsync-retries"):
                logging.error(lf('changelogs could not be processed '
                                 'completely - moving on...',
                                 files=list(map(os.path.basename, changes))))

                # Reset data counter on failure
                self.status.dec_value("data", self.files_in_batch)
                self.files_in_batch = 0
                self.datas_in_batch = set()

                if done:
                    xtl = (int(change.split('.')[-1]) - 1, 0)
                    self.upd_stime(xtl)
                    list(map(self.changelog_done_func, changes))
                    self.archive_and_purge_changelogs(changes)
                break
            # it's either entry_ops() or Rsync that failed to do it's
            # job. Mostly it's entry_ops() [which currently has a problem
            # of failing to create an entry but failing to return an errno]
            # Therefore we do not know if it's either Rsync or the freaking
            # entry_ops() that failed... so we retry the _whole_ changelog
            # again.
            # TODO: remove entry retries when it's gets fixed.
            logging.warn(lf('incomplete sync, retrying changelogs',
                            files=list(map(os.path.basename, changes))))

            # Reset the Data counter before Retry
            self.status.dec_value("data", self.files_in_batch)
            self.files_in_batch = 0
            self.init_fop_batch_stats()
            time.sleep(0.5)

        # Log the Skipped Entry ops range if any
        if self.skipped_entry_changelogs_first is not None and \
           self.skipped_entry_changelogs_last is not None:
            logging.info(lf("Skipping already processed entry ops",
                            from_changelog=self.skipped_entry_changelogs_first,
                            to_changelog=self.skipped_entry_changelogs_last,
                            num_changelogs=self.num_skipped_entry_changelogs))

        # Log Current batch details
        if changes:
            logging.info(
                lf("Entry Time Taken",
                   UNL=self.batch_stats["UNLINK"],
                   RMD=self.batch_stats["RMDIR"],
                   CRE=self.batch_stats["CREATE"],
                   MKN=self.batch_stats["MKNOD"],
                   MKD=self.batch_stats["MKDIR"],
                   REN=self.batch_stats["RENAME"],
                   LIN=self.batch_stats["LINK"],
                   SYM=self.batch_stats["SYMLINK"],
                   duration="%.4f" % self.batch_stats["ENTRY_SYNC_TIME"]))

            logging.info(
                lf("Data/Metadata Time Taken",
                   SETA=self.batch_stats["SETATTR"],
                   meta_duration="%.4f" % self.batch_stats["META_SYNC_TIME"],
                   SETX=self.batch_stats["SETXATTR"],
                   XATT=self.batch_stats["XATTROP"],
                   DATA=self.batch_stats["DATA"],
                   data_duration="%.4f" % (
                       time.time() - self.batch_stats["DATA_START_TIME"])))

            logging.info(
                lf("Batch Completed",
                   mode=self.name,
                   duration="%.4f" % (time.time() - self.batch_start_time),
                   changelog_start=changes[0].split(".")[-1],
                   changelog_end=changes[-1].split(".")[-1],
                   num_changelogs=len(changes),
                   stime=self.get_data_stime(),
                   entry_stime=self.get_entry_stime()))

    def upd_entry_stime(self, stime):
        self.slave.server.set_entry_stime(self.FLAT_DIR_HIERARCHY,
                                          self.uuid,
                                          stime)

    def upd_stime(self, stime, path=None):
        if not path:
            path = self.FLAT_DIR_HIERARCHY
        if not stime == URXTIME:
            self.sendmark(path, stime)

        # Update last_synced_time in status file based on stime
        # only update stime if stime xattr set to Brick root
        if path == self.FLAT_DIR_HIERARCHY:
            chkpt_time = gconf.getr("checkpoint")
            checkpoint_time = 0
            if chkpt_time is not None:
                checkpoint_time = int(chkpt_time)

            self.status.set_last_synced(stime, checkpoint_time)

    def update_worker_remote_node(self):
        node = rconf.args.resource_remote
        node_data = node.split("@")
        node = node_data[-1]
        remote_node_ip, _ = host_brick_split(node)
        self.status.set_slave_node(remote_node_ip)

    def changelogs_batch_process(self, changes):
        changelogs_batches = []
        current_size = 0
        for c in changes:
            si = os.lstat(c).st_size
            if (si + current_size) > gconf.get("changelog-batch-size"):
                # Create new batch if single Changelog file greater than
                # Max Size! or current batch size exceeds Max size
                changelogs_batches.append([c])
                current_size = si
            else:
                # Append to last batch, if No batches available Create one
                current_size += si
                if not changelogs_batches:
                    changelogs_batches.append([c])
                else:
                    changelogs_batches[-1].append(c)

        for batch in changelogs_batches:
            logging.debug(lf('processing changes',
                             batch=batch))
            self.process(batch)

    def crawl(self):
        self.status.set_worker_crawl_status("Changelog Crawl")
        changes = []
        # get stime (from the brick) and purge changelogs
        # that are _historical_ to that time.
        data_stime = self.get_data_stime()

        libgfchangelog.scan()
        self.crawls += 1
        changes = libgfchangelog.getchanges()
        if changes:
            if data_stime:
                logging.info(lf("slave's time",
                                stime=data_stime))
                processed = [x for x in changes
                             if int(x.split('.')[-1]) < data_stime[0]]
                for pr in processed:
                    logging.debug(
                        lf('skipping already processed change',
                           changelog=os.path.basename(pr)))
                    self.changelog_done_func(pr)
                    changes.remove(pr)
                self.archive_and_purge_changelogs(processed)

        self.changelogs_batch_process(changes)

    def register(self, register_time, status):
        self.sleep_interval = gconf.get("change-interval")
        self.changelog_done_func = libgfchangelog.done
        self.tempdir = self.setup_working_dir()
        self.processed_changelogs_dir = os.path.join(self.tempdir,
                                                     ".processed")
        self.name = "live_changelog"
        self.status = status


class GMasterChangeloghistoryMixin(GMasterChangelogMixin):
    def register(self, register_time, status):
        self.changelog_register_time = register_time
        self.history_crawl_start_time = register_time
        self.changelog_done_func = libgfchangelog.history_done
        self.history_turns = 0
        self.tempdir = self.setup_working_dir()
        self.processed_changelogs_dir = os.path.join(self.tempdir,
                                                     ".history/.processed")
        self.name = "history_changelog"
        self.status = status

    def crawl(self):
        self.history_turns += 1
        self.status.set_worker_crawl_status("History Crawl")
        data_stime = self.get_data_stime()

        end_time = int(time.time())

        #as start of historical crawl marks Geo-rep worker restart
        if gconf.get("ignore-deletes"):
            logging.info(lf('ignore-deletes config option is set',
                         stime=data_stime))

        logging.info(lf('starting history crawl',
                        turns=self.history_turns,
                        stime=data_stime,
                        etime=end_time,
                        entry_stime=self.get_entry_stime()))

        if not data_stime or data_stime == URXTIME:
            raise NoStimeAvailable()

        # Changelogs backend path is hardcoded as
        # <BRICK_PATH>/.glusterfs/changelogs, if user configured to different
        # location then consuming history will not work(Known issue as of now)
        changelog_path = os.path.join(rconf.args.local_path,
                                      ".glusterfs/changelogs")
        ret, actual_end = libgfchangelog.history_changelog(
            changelog_path,
            data_stime[0],
            end_time,
            gconf.get("sync-jobs"))

        # scan followed by getchanges till scan returns zero.
        # history_scan() is blocking call, till it gets the number
        # of changelogs to process. Returns zero when no changelogs
        # to be processed. returns positive value as number of changelogs
        # to be processed, which will be fetched using
        # history_getchanges()
        while libgfchangelog.history_scan() > 0:
            self.crawls += 1

            changes = libgfchangelog.history_getchanges()
            if changes:
                if data_stime:
                    logging.info(lf("slave's time",
                                    stime=data_stime))
                    processed = [x for x in changes
                                 if int(x.split('.')[-1]) < data_stime[0]]
                    for pr in processed:
                        logging.debug(lf('skipping already processed change',
                                         changelog=os.path.basename(pr)))
                        self.changelog_done_func(pr)
                        changes.remove(pr)

            self.changelogs_batch_process(changes)

        history_turn_time = int(time.time()) - self.history_crawl_start_time

        logging.info(lf('finished history crawl',
                        endtime=actual_end,
                        stime=self.get_data_stime(),
                        entry_stime=self.get_entry_stime()))

        # If TS returned from history_changelog is < register_time
        # then FS crawl may be required, since history is only available
        # till TS returned from history_changelog
        if actual_end < self.changelog_register_time:
            if self.history_turns < 2:
                sleep_time = 1
                if history_turn_time < CHANGELOG_ROLLOVER_TIME:
                    sleep_time = CHANGELOG_ROLLOVER_TIME - history_turn_time
                time.sleep(sleep_time)
                self.history_crawl_start_time = int(time.time())
                self.crawl()
            else:
                # This exception will be caught in resource.py and
                # fallback to xsync for the small gap.
                raise PartialHistoryAvailable(str(actual_end))


class GMasterXsyncMixin(GMasterChangelogMixin):

    """
    This crawl needs to be xtime based (as of now
    it's not. this is because we generate CHANGELOG
    file during each crawl which is then processed
    by process_change()).
    For now it's used as a one-shot initial sync
    mechanism and only syncs directories, regular
    files, hardlinks and symlinks.
    """

    XSYNC_MAX_ENTRIES = 1 << 13

    def register(self, register_time=None, status=None):
        self.status = status
        self.counter = 0
        self.comlist = []
        self.stimes = []
        self.sleep_interval = 60
        self.tempdir = self.setup_working_dir()
        logging.info(lf('Working dir',
                        path=self.tempdir))
        self.tempdir = os.path.join(self.tempdir, 'xsync')
        self.processed_changelogs_dir = self.tempdir
        self.name = "xsync"
        try:
            os.makedirs(self.tempdir)
        except OSError:
            ex = sys.exc_info()[1]
            if ex.errno == EEXIST and os.path.isdir(self.tempdir):
                pass
            else:
                raise
        # Purge stale unprocessed xsync changelogs
        for f in os.listdir(self.tempdir):
            if f.startswith("XSYNC-CHANGELOG"):
                os.remove(os.path.join(self.tempdir, f))


    def crawl(self):
        """
        event dispatcher thread

        this thread dispatches either changelog or synchronizes stime.
        additionally terminates itself on receiving a 'finale' event
        """
        def Xsyncer():
            self.Xcrawl()
        t = Thread(target=Xsyncer)
        t.start()
        logging.info(lf('starting hybrid crawl',
                        stime=self.get_data_stime()))
        self.status.set_worker_crawl_status("Hybrid Crawl")
        while True:
            try:
                item = self.comlist.pop(0)
                if item[0] == 'finale':
                    logging.info(lf('finished hybrid crawl',
                                    stime=self.get_data_stime()))
                    break
                elif item[0] == 'xsync':
                    logging.info(lf('processing xsync changelog',
                                    path=item[1]))
                    self.process([item[1]], 0)
                    self.archive_and_purge_changelogs([item[1]])
                elif item[0] == 'stime':
                    logging.debug(lf('setting slave time',
                                     time=item[1]))
                    self.upd_stime(item[1][1], item[1][0])
                else:
                    logging.warn(lf('unknown tuple in comlist',
                                    entry=item))
            except IndexError:
                time.sleep(1)

    def write_entry_change(self, prefix, data=[]):
        if not getattr(self, "fh", None):
            self.open()

        self.fh.write("%s %s\n" % (prefix, ' '.join(data)))

    def open(self):
        try:
            self.xsync_change = os.path.join(
                self.tempdir, 'XSYNC-CHANGELOG.' + str(int(time.time())))
            self.fh = open(self.xsync_change, 'w')
        except IOError:
            raise

    def close(self):
        if getattr(self, "fh", None):
            self.fh.flush()
            os.fsync(self.fh.fileno())
            self.fh.close()
            self.fh = None

    def fname(self):
        return self.xsync_change

    def put(self, mark, item):
        self.comlist.append((mark, item))

    def sync_xsync(self, last):
        """schedule a processing of changelog"""
        self.close()
        if self.counter > 0:
            self.put('xsync', self.fname())
        self.counter = 0
        if not last:
            time.sleep(1)  # make sure changelogs are 1 second apart

    def sync_stime(self, stime=None, last=False):
        """schedule a stime synchronization"""
        if stime:
            self.put('stime', stime)
        if last:
            self.put('finale', None)

    def sync_done(self, stime=[], last=False):
        self.sync_xsync(last)
        if stime:
            # Send last as True only for last stime entry
            for st in stime[:-1]:
                self.sync_stime(st, False)

            if stime and stime[-1]:
                self.sync_stime(stime[-1], last)

    def is_sticky(self, path, mo):
        """check for DHTs linkto sticky bit file"""
        sticky = False
        if mo & 0o1000:
            sticky = self.master.server.linkto_check(path)
        return sticky

    def Xcrawl(self, path='.', xtr_root=None):
        """
        generate a CHANGELOG file consumable by process_change.

        slave's xtime (stime) is _cached_ for comparisons across
        the filesystem tree, but set after directory synchronization.
        """
        if path == '.':
            self.crawls += 1
        if not xtr_root:
            # get the root stime and use it for all comparisons
            xtr_root = self.xtime('.', self.slave)
            if isinstance(xtr_root, int):
                if xtr_root != ENOENT:
                    logging.warn(lf("slave cluster not returning the "
                                    "xtime for root",
                                    error=xtr_root))
                xtr_root = self.minus_infinity
        xtl = self.xtime(path)
        if isinstance(xtl, int):
            logging.warn("master cluster's xtime not found")
        xtr = self.xtime(path, self.slave)
        if isinstance(xtr, int):
            if xtr != ENOENT:
                logging.warn(lf("slave cluster not returning the "
                                "xtime for dir",
                                path=path,
                                error=xtr))
            xtr = self.minus_infinity
        xtr = max(xtr, xtr_root)
        zero_zero = (0, 0)
        if xtr_root == zero_zero:
            xtr = self.minus_infinity
        if not self.need_sync(path, xtl, xtr):
            if path == '.':
                self.sync_done([(path, xtl)], True)
            return
        self.xtime_reversion_hook(path, xtl, xtr)
        logging.debug("entering " + path)
        dem = self.master.server.entries(path)
        pargfid = self.master.server.gfid(path)
        if isinstance(pargfid, int):
            logging.warn(lf('skipping directory',
                            path=path))
        for e in dem:
            bname = e
            e = os.path.join(path, e)
            xte = self.xtime(e)
            if isinstance(xte, int):
                logging.warn(lf("irregular xtime",
                                path=e,
                                error=errno.errorcode[xte]))
                continue
            if not self.need_sync(e, xte, xtr):
                continue
            st = self.master.server.lstat(e)
            if isinstance(st, int):
                logging.warn(lf('got purged in the interim',
                                path=e))
                continue
            if self.is_sticky(e, st.st_mode):
                logging.debug(lf('ignoring sticky bit file',
                                 path=e))
                continue
            gfid = self.master.server.gfid(e)
            if isinstance(gfid, int):
                logging.warn(lf('skipping entry',
                                path=e))
                continue
            mo = st.st_mode
            self.counter += 1 if ((stat.S_ISDIR(mo) or
                                   stat.S_ISLNK(mo) or
                                   stat.S_ISREG(mo))) else 0
            if self.counter == self.XSYNC_MAX_ENTRIES:
                self.sync_done(self.stimes, False)
                self.stimes = []
            if stat.S_ISDIR(mo):
                self.write_entry_change("E",
                                        [gfid, 'MKDIR', str(mo),
                                         str(0), str(0), escape_space_newline(
                                             os.path.join(pargfid, bname))])
                self.write_entry_change("M", [gfid, "SETATTR", str(st.st_uid),
                                              str(st.st_gid), str(st.st_mode),
                                              str(st.st_atime),
                                              str(st.st_mtime)])
                self.Xcrawl(e, xtr_root)
                stime_to_update = xte
                # Live Changelog Start time indicates that from that time
                # onwards Live changelogs are available. If we update stime
                # greater than live_changelog_start time then Geo-rep will
                # skip those changelogs as already processed. But Xsync
                # actually failed to sync the deletes and Renames. Update
                # stime as min(Live_changelogs_time, Actual_stime) When it
                # switches to Changelog mode, it syncs Deletes and Renames.
                if self.live_changelog_start_time:
                    stime_to_update = min(self.live_changelog_start_time, xte)
                self.stimes.append((e, stime_to_update))
            elif stat.S_ISLNK(mo):
                self.write_entry_change(
                    "E", [gfid, 'SYMLINK', escape_space_newline(
                        os.path.join(pargfid, bname))])
            elif stat.S_ISREG(mo):
                nlink = st.st_nlink
                nlink -= 1  # fixup backend stat link count
                # if a file has a hardlink, create a Changelog entry as
                # 'LINK' so the slave side will decide if to create the
                # new entry, or to create link.
                if nlink == 1:
                    self.write_entry_change("E",
                                            [gfid, 'MKNOD', str(mo),
                                             str(0), str(0),
                                             escape_space_newline(
                                                 os.path.join(
                                                     pargfid, bname))])
                else:
                    self.write_entry_change(
                        "E", [gfid, 'LINK', escape_space_newline(
                            os.path.join(pargfid, bname))])
                self.write_entry_change("D", [gfid])
        if path == '.':
            stime_to_update = xtl
            if self.live_changelog_start_time:
                stime_to_update = min(self.live_changelog_start_time, xtl)
            self.stimes.append((path, stime_to_update))
            self.sync_done(self.stimes, True)


class BoxClosedErr(Exception):
    pass


class PostBox(list):

    """synchronized collection for storing things thought of as "requests" """

    def __init__(self, *a):
        list.__init__(self, *a)
        # too bad Python stdlib does not have read/write locks...
        # it would suffivce to grab the lock in .append as reader, in .close as
        # writer
        self.lever = Condition()
        self.open = True
        self.done = False

    def wait(self):
        """wait on requests to be processed"""
        self.lever.acquire()
        if not self.done:
            self.lever.wait()
        self.lever.release()
        return self.result

    def wakeup(self, data):
        """wake up requestors with the result"""
        self.result = data
        self.lever.acquire()
        self.done = True
        self.lever.notifyAll()
        self.lever.release()

    def append(self, e):
        """post a request"""
        self.lever.acquire()
        if not self.open:
            raise BoxClosedErr
        list.append(self, e)
        self.lever.release()

    def close(self):
        """prohibit the posting of further requests"""
        self.lever.acquire()
        self.open = False
        self.lever.release()


class Syncer(object):

    """a staged queue to relay rsync requests to rsync workers

    By "staged queue" its meant that when a consumer comes to the
    queue, it takes _all_ entries, leaving the queue empty.
    (I don't know if there is an official term for this pattern.)

    The queue uses a PostBox to accumulate incoming items.
    When a consumer (rsync worker) comes, a new PostBox is
    set up and the old one is passed on to the consumer.

    Instead of the simplistic scheme of having one big lock
    which synchronizes both the addition of new items and
    PostBox exchanges, use a separate lock to arbitrate consumers,
    and rely on PostBox's synchronization mechanisms take
    care about additions.

    There is a corner case racy situation, producers vs. consumers,
    which is not handled by this scheme: namely, when the PostBox
    exchange occurs in between being passed to the producer for posting
    and the post placement. But that's what Postbox.close is for:
    such a posting will find the PostBox closed, in which case
    the producer can re-try posting against the actual PostBox of
    the queue.

    To aid accumlation of items in the PostBoxen before grabbed
    by an rsync worker, the worker goes to sleep a bit after
    each completed syncjob.
    """

    def __init__(self, slave, sync_engine, resilient_errnos=[]):
        """spawn worker threads"""
        self.log_err = False
        self.slave = slave
        self.lock = Lock()
        self.pb = PostBox()
        self.sync_engine = sync_engine
        self.errnos_ok = resilient_errnos
        for i in range(gconf.get("sync-jobs")):
            t = Thread(target=self.syncjob, args=(i + 1, ))
            t.start()

    def syncjob(self, job_id):
        """the life of a worker"""
        while True:
            pb = None
            while True:
                self.lock.acquire()
                if self.pb:
                    pb, self.pb = self.pb, PostBox()
                self.lock.release()
                if pb:
                    break
                time.sleep(0.5)
            pb.close()
            start = time.time()
            po = self.sync_engine(pb, self.log_err)
            logging.info(lf("Sync Time Taken",
                            job=job_id,
                            num_files=len(pb),
                            return_code=po.returncode,
                            duration="%.4f" % (time.time() - start)))

            if po.returncode == 0:
                ret = (True, 0)
            elif po.returncode in self.errnos_ok:
                ret = (False, po.returncode)
            else:
                po.errfail()
            pb.wakeup(ret)

    def add(self, e):
        while True:
            pb = self.pb
            try:
                pb.append(e)
                return pb
            except BoxClosedErr:
                pass

    def enable_errorlog(self):
        self.log_err = True

    def disable_errorlog(self):
        self.log_err = False



# -*- coding: utf-8 -*-
# Reexporting the utility funcs and classes
from .cliutils import (runcli,
                       sync_file_to_peers,
                       execute_in_peers,
                       execute,
                       node_output_ok,
                       node_output_notok,
                       output_error,
                       oknotok,
                       yesno,
                       get_node_uuid,
                       Cmd,
                       GlusterCmdException,
                       set_common_args_func)


# This will be useful when `from cliutils import *`
__all__ = ["runcli",
           "sync_file_to_peers",
           "execute_in_peers",
           "execute",
           "node_output_ok",
           "node_output_notok",
           "output_error",
           "oknotok",
           "yesno",
           "get_node_uuid",
           "Cmd",
           "GlusterCmdException",
           "set_common_args_func"]
